import numpy as np
import pytest

try:
    import healpy as hp
    has_healpy = True
except ImportError:
    has_healpy = False

import hpgeom


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10])
def test_ring_to_nest_allpix(nside):
    """Test ring_to_nest for all pixels."""
    ring_pix = np.arange(12*nside*nside)

    nest_pix_hpgeom = hpgeom.ring_to_nest(nside, ring_pix)
    nest_pix_healpy = hp.ring2nest(nside, ring_pix)

    np.testing.assert_array_equal(nest_pix_hpgeom, nest_pix_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside", [2**15, 2**20, 2**25])
def test_ring_to_nest_samplepix(nside):
    """Test ring_to_nest for sampled pixels."""
    np.random.seed(12345)

    ring_pix = np.random.randint(low=0, high=12*nside*nside - 1, size=1_000_000)

    nest_pix_hpgeom = hpgeom.ring_to_nest(nside, ring_pix)
    nest_pix_healpy = hp.ring2nest(nside, ring_pix)

    np.testing.assert_array_equal(nest_pix_hpgeom, nest_pix_healpy)


@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10, 2**15, 2**20, 2**25, 2**29])
def test_ring_to_nest_bad_pix(nside):
    """Test ring_to_nest errors when given bad pixel"""

    with pytest.raises(ValueError, match=r"Pixel value .* out of range"):
        hpgeom.ring_to_nest(nside, -1)

    with pytest.raises(ValueError, match=r"Pixel value .* out of range"):
        hpgeom.ring_to_nest(nside, 12*nside*nside)


def test_ring_to_nest_bad_nside():
    """test ring_to_nest errors when given a bad nside."""

    with pytest.raises(ValueError, match=r"nside .* must be positive"):
        hpgeom.ring_to_nest(-10, 100)

    with pytest.raises(ValueError, match=r"nside .* must be power of 2"):
        hpgeom.ring_to_nest(1020, 100)



import numpy as np
import pytest

try:
    import healpy as hp
    has_healpy = True
except ImportError:
    has_healpy = False

import hpgeom


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_angle_to_vector():
    """Test angle_to_vector."""
    np.random.seed(12345)

    lon = np.random.uniform(0, 360.0, 1_000_000)
    lat = np.random.uniform(-90.0, 90.0, 1_000_000)

    vec_hpgeom = hpgeom.angle_to_vector(lon, lat)
    vec_healpy = hp.ang2vec(lon, lat, lonlat=True)

    np.testing.assert_array_almost_equal(vec_hpgeom, vec_healpy)

    vec_hpgeom = hpgeom.angle_to_vector(float(lon[0]), float(lat[0]))
    vec_healpy = hp.ang2vec(float(lon[0]), float(lat[0]), lonlat=True)

    assert vec_hpgeom.shape == (3, )
    np.testing.assert_array_almost_equal(vec_hpgeom, vec_healpy)

    theta, phi = hpgeom.lonlat_to_thetaphi(lon, lat)

    vec_hpgeom = hpgeom.angle_to_vector(theta, phi, lonlat=False)
    vec_healpy = hp.ang2vec(theta, phi)

    np.testing.assert_array_almost_equal(vec_hpgeom, vec_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_vector_to_angle():
    """Test angle_to_vector."""
    np.random.seed(12345)

    # Does not need to be normalized.
    vec = np.zeros((1_000_000, 3))
    vec[:, 0] = np.random.uniform(-1, 1, 1_000_000)
    vec[:, 1] = np.random.uniform(-1, 1, 1_000_000)
    vec[:, 2] = np.random.uniform(-1, 1, 1_000_000)

    lon_hpgeom, lat_hpgeom = hpgeom.vector_to_angle(vec)
    lon_healpy, lat_healpy = hp.vec2ang(vec, lonlat=True)

    np.testing.assert_array_almost_equal(lon_hpgeom, lon_healpy)
    np.testing.assert_array_almost_equal(lat_hpgeom, lat_healpy)

    theta_hpgeom, phi_hpgeom = hpgeom.vector_to_angle(vec, lonlat=False)
    theta_healpy, phi_healpy = hp.vec2ang(vec)

    np.testing.assert_array_almost_equal(theta_hpgeom, theta_healpy)
    np.testing.assert_array_almost_equal(phi_hpgeom, phi_healpy)


def test_angle_to_vector_badinput():
    """Test angle_to_vector."""

    with pytest.raises(ValueError, match=r"Latitude out of range"):
        hpgeom.angle_to_vector(0.0, 100.0, lonlat=True)

    with pytest.raises(ValueError, match=r"Co-latitude \(theta\) out of range"):
        hpgeom.angle_to_vector(2*np.pi, 0.0, lonlat=False)



import numpy as np
import pytest

try:
    import healpy as hp
    has_healpy = True
except ImportError:
    has_healpy = False

import hpgeom


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10, 2**20, 2**29])
def test_nside_to_npixel(nside):
    """Test nside_to_npixel."""
    npixel_hpgeom = hpgeom.nside_to_npixel(nside)
    npixel_healpy = hp.nside2npix(nside)

    assert (npixel_hpgeom == npixel_healpy)

    npixel_hpgeom = hpgeom.nside_to_npixel(np.array([nside, nside]))
    npixel_healpy = hp.nside2npix(np.array([nside, nside]))

    np.testing.assert_array_equal(npixel_hpgeom, npixel_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10, 2**20, 2**29])
def test_npixel_to_nside(nside):
    """Test npixel_to_nside."""
    npixel = 12*nside*nside

    nside_hpgeom = hpgeom.npixel_to_nside(npixel)
    nside_healpy = hp.npix2nside(npixel)

    assert (nside_hpgeom == nside_healpy)

    nside_hpgeom = hpgeom.npixel_to_nside(np.array([npixel, npixel]))

    np.testing.assert_array_equal(nside_hpgeom, nside)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10, 2**20, 2**29])
def test_nside_to_pixel_area(nside):
    """Test nside_to_pixel_area."""
    pixel_area_hpgeom = hpgeom.nside_to_pixel_area(nside)
    pixel_area_healpy = hp.nside2pixarea(nside, degrees=True)

    assert (pixel_area_hpgeom == pixel_area_healpy)

    pixel_area_hpgeom = hpgeom.nside_to_pixel_area(nside, degrees=False)
    pixel_area_healpy = hp.nside2pixarea(nside, degrees=False)

    assert (pixel_area_hpgeom == pixel_area_healpy)

    pixel_area_hpgeom_arr = hpgeom.nside_to_pixel_area(np.array([nside, nside]))

    np.testing.assert_array_equal(pixel_area_hpgeom_arr, hpgeom.nside_to_pixel_area(nside))


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10, 2**20, 2**29])
def test_nside_to_resolution(nside):
    """Test nside_to_resolution."""
    resolution_hpgeom = hpgeom.nside_to_resolution(nside)
    resolution_healpy = np.rad2deg(hp.nside2resol(nside))

    assert (resolution_hpgeom == resolution_healpy)

    resolution_hpgeom = hpgeom.nside_to_resolution(nside, units='radians')
    resolution_healpy = hp.nside2resol(nside)

    assert (resolution_hpgeom == resolution_healpy)

    resolution_hpgeom = hpgeom.nside_to_resolution(nside, units='arcminutes')
    resolution_healpy = np.rad2deg(hp.nside2resol(nside))*60.

    assert (resolution_hpgeom == resolution_healpy)

    resolution_hpgeom = hpgeom.nside_to_resolution(nside, units='arcseconds')
    resolution_healpy = np.rad2deg(hp.nside2resol(nside))*60.*60.

    assert (resolution_hpgeom == resolution_healpy)

    resolution_hpgeom_arr = hpgeom.nside_to_resolution(np.array([nside, nside]))

    np.testing.assert_array_almost_equal(resolution_hpgeom_arr, hpgeom.nside_to_resolution(nside))


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10, 2**20, 2**29])
def test_nside_to_order(nside):
    """Test nside_to_order."""
    order_hpgeom = hpgeom.nside_to_order(nside)
    order_healpy = hp.nside2order(nside)

    assert (order_hpgeom == order_healpy)

    order_hpgeom_arr = hpgeom.nside_to_order(np.array([nside, nside]))

    np.testing.assert_array_equal(order_hpgeom_arr, order_hpgeom)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("order", [0, 5, 10, 20, 29])
def test_order_to_nside(order):
    """Test order_to_nside."""
    nside_hpgeom = hpgeom.order_to_nside(order)
    nside_healpy = hp.order2nside(order)

    assert (nside_hpgeom == nside_healpy)

    nside_hpgeom_arr = hpgeom.order_to_nside(np.array([order, order]))

    np.testing.assert_array_equal(nside_hpgeom_arr, nside_hpgeom)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside", [2**5, 2**10])
@pytest.mark.parametrize("ring_to_nest", [False, True])
def test_reorder(nside, ring_to_nest):
    """Test reorder."""
    npix = hpgeom.nside_to_npixel(nside)
    map_in = np.arange(npix)

    map_out_hpgeom = hpgeom.reorder(map_in, ring_to_nest=ring_to_nest)

    if (ring_to_nest):
        map_out_healpy = hp.reorder(map_in, r2n=True)
    else:
        map_out_healpy = hp.reorder(map_in, n2r=True)

    np.testing.assert_array_equal(map_out_hpgeom, map_out_healpy)


def test_bad_nsides():
    """Test raising when bad nsides given."""

    with pytest.raises(ValueError, match=r"Illegal nside value"):
        hpgeom.nside_to_npixel(-1)

    with pytest.raises(ValueError, match=r"Illegal nside value"):
        hpgeom.nside_to_npixel(2**30)

    with pytest.raises(ValueError, match=r"Illegal nside value"):
        hpgeom.nside_to_npixel(np.array([1024, -1]))

    with pytest.raises(ValueError, match=r"Illegal npixel"):
        hpgeom.npixel_to_nside(-1)

    with pytest.raises(ValueError, match=r"Illegal npixel"):
        hpgeom.npixel_to_nside(10000)

    with pytest.raises(ValueError, match=r"Illegal npixel"):
        hpgeom.npixel_to_nside(np.array([-1, 12*1024*1024]))

    with pytest.raises(ValueError, match=r"Illegal nside value"):
        hpgeom.nside_to_pixel_area(2**30)

    with pytest.raises(ValueError, match=r"Illegal nside value"):
        hpgeom.nside_to_pixel_area(np.array([-1, 1024]))

    with pytest.raises(ValueError, match=r"Illegal nside value"):
        hpgeom.nside_to_resolution(2**30)

    with pytest.raises(ValueError, match=r"Illegal nside value"):
        hpgeom.nside_to_resolution(np.array([-1, 1024]))



import numpy as np


__all__ = ['match_arrays']


def match_arrays(arr1, arr2):
    """Match arrays and return indices.

    Parameters
    ----------
    arr1 : `np.ndarray`
    arr2 : `np.ndarray`

    Returns
    -------
    sub1 : `np.ndarray`
        Indices of arr1 which match arr2
    sub2 : `np.ndarray`
        Indices of arr2 which match arr1
    """
    sub1 = np.searchsorted(arr1, arr2)
    sub1[sub1 == arr1.size] = arr1.size - 1
    sub2, = (arr1[sub1] == arr2).nonzero()
    sub1 = sub1[sub2]

    return sub1, sub2



import numpy as np

from ._hpgeom import (
    angle_to_pixel,
    pixel_to_angle,
    query_circle,
    query_polygon,
    query_ellipse,
    query_box,
    nest_to_ring,
    ring_to_nest,
    vector_to_pixel,
    pixel_to_vector,
    boundaries,
    neighbors,
    max_pixel_radius,
)

__all__ = [
    'thetaphi_to_lonlat',
    'lonlat_to_thetaphi',
    'angle_to_pixel',
    'pixel_to_angle',
    'query_circle',
    'query_circle_vec',
    'query_polygon',
    'query_polygon_vec',
    'query_ellipse',
    'query_box',
    'lonlat_to_thetaphi',
    'nest_to_ring',
    'ring_to_nest',
    'nside_to_npixel',
    'npixel_to_nside',
    'nside_to_pixel_area',
    'nside_to_resolution',
    'nside_to_order',
    'order_to_nside',
    'angle_to_vector',
    'vector_to_angle',
    'vector_to_pixel',
    'pixel_to_vector',
    'boundaries',
    'neighbors',
    'max_pixel_radius',
    'reorder',
    'UNSEEN',
]

UNSEEN = -1.6375e+30
max_nside = 1 << 29


def lonlat_to_thetaphi(lon, lat, degrees=True):
    """Convert longitude/latitude to theta/phi.

    Parameters
    ----------
    lon, lat : `float` or `np.ndarray` (N,)
        Longitude/latitude.
    degrees : `bool`, optional
        If True, longitude and latitude will be in degrees.

    Returns
    -------
    theta, phi : `float` or `np.ndarray` (N,)
        Co-latitude (theta) and longitude (phi) in radians.
    """
    if degrees:
        lon_ = np.deg2rad(lon % 360.)
        lat_ = np.deg2rad(lat)
    else:
        lon_ = lon % (2.*np.pi)
        lat_ = lat

    if np.any((lat_ < -np.pi/2.) | (lat_ > np.pi/2.)):
        raise ValueError("Latitude out of range.")

    theta = np.pi/2. - lat_
    phi = lon_

    return theta, phi


def thetaphi_to_lonlat(theta, phi, degrees=True):
    """Convert theta/phi to longitude/latitude.

    Parameters
    ----------
    theta, phi : `float` or `np.ndarray` (N,)
        Co-latitude (theta) and longitude (phi) in radians.
    degrees : `bool`, optional
        If True, longitude and latitude will be in degrees.

    Returns
    -------
    lon, lat : `float` or `np.ndarray` (N,)
        Longitude/latitude.
    """
    lon = phi
    lat = -(theta - np.pi/2.)
    if (degrees):
        lon = np.rad2deg(lon)
        lat = np.rad2deg(lat)

    return lon, lat


def query_circle_vec(nside, vec, radius, inclusive=False, fact=4, nest=True):
    """Returns pixels whose centers lie within the circle defined by vec
    and radius (in radians) if inclusive is False, or which overlap with
    this circle (if inclusive is True).

    Parameters
    ----------
    nside : `int`
        HEALPix nside. Must be power of 2 for nest ordering.
    vec : iterable with 3 elements, `float`
        The coordinates of the unit vector defining the circle center.
    radius : `float`
        The radius (in radians) of the circle.
    inclusive : `bool`, optional
        If False, return the exact set of pixels whose pixel centers lie
        within the circle. If True, return all pixels that overlap with
        the disk. This is an approximation and may return a few extra
        pixels.
    fact : `int`, optional
        Only used when inclusive=True. The overlap test is performed at
        a resolution fact*nside. For nest ordering, fact must be a power
        of 2, and nside*fact must always be <= 2**29.  For ring ordering
        fact may be any positive integer.
    nest : `bool`, optional
        If True, use nest ordering.

    Returns
    -------
    pixels : `np.ndarray` (N,)
        Array of pixels (`np.int64`) which cover the circle.
    """
    if len(vec) != 3:
        raise ValueError("vec must be 3 elements.")

    theta, phi = vector_to_angle(vec, lonlat=False)

    return query_circle(nside, theta, phi, radius, inclusive=inclusive, fact=fact, nest=nest, lonlat=False)


def query_polygon_vec(nside, vertices, inclusive=False, fact=4, nest=True):
    """Returns pixels whose centers lie within the convex polygon defined
    by vertices (if inclusive is False), or which overlap with the polygon
    (if inclusive is True).

    Parameters
    ----------
    nside : `int`
        HEALPix nside. Must be power of 2 for nest ordering.
    vertices : `np.ndarray` (N, 3)
        Vertex array containing the vertices of the polygon.
    inclusive : `bool`, optional
        If False, return the exact set of pixels whose pixel centers lie
        within the polygon; if True, return all pixels that overlap with the polygon,
        and maybe a few more.
    fact : `int`, optional
        Only used when inclusive=True. The overlapping test will be done at
        the resolution fact*nside. For NESTED ordering, fact must be a power of 2, less than 2**30,
        else it can be any positive integer.
    nest: `bool`, optional
        Use nest ordering scheme?
    """
    theta, phi = vector_to_angle(vertices, lonlat=False)

    return query_polygon(nside, theta, phi, inclusive=inclusive, fact=fact, nest=nest, lonlat=False)


def nside_to_npixel(nside):
    """Return the number of pixels given an nside.

    Parameters
    ----------
    nside : `int` or `np.ndarray` (N,)
        HEALPix nside.

    Returns
    -------
    npixel : `int` or `np.ndarray` (N,)
        Number of pixels associated with that nside.
    """
    _nside = np.atleast_1d(nside)
    if np.any((_nside < 0) | (_nside > max_nside)):
        raise ValueError("Illegal nside value (must be 0 <= nside <= 2**29)")
    return 12*nside*nside


def npixel_to_nside(npixel):
    """Return the nside given a number of pixels.

    Parameters
    ----------
    npixel : `int` or `np.ndarray` (N,)
        Number of pixels.

    Returns
    -------
    nside : `int` or `np.ndarray` (N,)
        HEALPix nside associated with that number of pixels.
    """
    if (np.any(npixel <= 0)):
        raise ValueError("Illegal npixel (must be positive)")
    nside = np.sqrt(np.atleast_1d(npixel)/12.0)
    if np.any(nside != np.floor(nside)):
        raise ValueError("Illegal npixel (it must be 12*nside*nside)")

    if len(nside) == 1:
        return int(nside[0])
    else:
        return nside.astype(np.int64)


def nside_to_pixel_area(nside, degrees=True):
    """Return the pixel area given an nside in square degrees or square radians.

    Parameters
    ----------
    nside : `int`
        HEALPix nside parameter.
    degrees : `bool`, optional
        Return area in square degrees? Otherwise square radians.

    Returns
    -------
    pixel_area : `float`
        Pixel area in square degrees or square radians.
    """
    pixel_area = 4*np.pi/nside_to_npixel(nside)

    if degrees:
        value = np.rad2deg(np.rad2deg(pixel_area))
    else:
        value = pixel_area

    return value


def nside_to_resolution(nside, units='degrees'):
    """Return the approximate resolution (pixel size in radians, arcseconds, arcminutes,
    or degrees) given an nside.

    Resolution is just the square root of the pixel area, which is an approximation
    given the varying pixel shapes.

    Parameters
    ----------
    nside : `int`
        HEALPix nside parameter.
    units : `str`, optional
        Units to return.  Valid options are ``radians``, ``degrees``, ``arcminutes``,
        ``arcseconds``.

    Returns
    -------
    resolution : `float`
        Approximate pixel size in specified units.
    """
    resolution = np.sqrt(nside_to_pixel_area(nside, degrees=False))

    if units == 'radians':
        value = resolution
    elif units == 'degrees':
        value = np.rad2deg(resolution)
    elif units == 'arcminutes':
        value = np.rad2deg(resolution)*60.
    elif units == 'arcseconds':
        value = np.rad2deg(resolution)*60.*60.
    else:
        raise ValueError("Invalid units.  Must be radians, degrees, arcminutes, or arcseconds.")

    return value


def nside_to_order(nside):
    """Return the resolution order for a given nside.

    Parameters
    ----------
    nside : `int` or `np.ndarray` (N,)
        HEALPix nside parameter.

    Returns
    -------
    order : `int` or `np.ndarray` (N,)
        Order corresponding to given nside, such that nside = 2**order.

    Raises
    ------
    ValueError if nside is not valid (must be power of 2 and less than 2**30).
    """
    _nside = np.atleast_1d(nside)
    if np.any((_nside <= 0) | (_nside > max_nside) | ((_nside & (_nside - 1)) != 0)):
        raise ValueError("nside must be postive power of 2, and less than 2**30")

    if (hasattr(nside, '__len__')):
        return np.round(np.log2(nside)).astype(np.int64)
    else:
        return int(np.round(np.log2(nside)))


def order_to_nside(order):
    """Return the nside for a given order.

    Parameters
    ----------
    order : `int` or `np.ndarray` (N,)
        Resolution order.  Will raise ValueError if order is not valid
        (must be 0 to 29 inclusive).

    Returns
    -------
    nside : `int` or `np.ndarray` (N,)
        HEALPix nside corresponding to given order, such that nside = 2**order.
    """
    if np.any((order != np.int64(order)) | (order < 0) | (order > 29)):
        raise ValueError("Order must be integer, 0<=order<=29.")

    return 2**order


def _check_theta_phi(theta, phi):
    """Check that theta/phi are valid.

    Parameters
    ----------
    theta : `float` or `np.ndarray` (N,)
        Co-latitude in radians.
    phi : `float` or `np.ndarray` (N,)
        Longitude in radians.

    Raises
    ------
    ValueError if theta or phi is invalid.
    """
    _theta = np.atleast_1d(theta)
    _phi = np.atleast_1d(phi)
    if np.any((_theta < 0.0) | (_theta > np.pi)):
        raise ValueError("Co-latitude (theta) out of range.")
    if np.any((_phi < 0.0) | (_phi > 2*np.pi)):
        raise ValueError("Longitude (phi) out of range.")


def angle_to_vector(a, b, lonlat=True, degrees=True):
    """Convert angles to cartesion (x, y, z) unit vectors.

    Parameters
    ----------
    a, b : `float` or `np.ndarray` (N,)
        Longitude/latitude (if lonlat=True) or Co-latitude(theta)/longitude(phi)
        (if lonlat=False). Longitude/latitude will be in degrees if degrees=True
        and in radians if degrees=False. Theta/phi are always in radians.
    lonlat : `bool`, optional
        Use longitude/latitude instead of co-latitude/longitude (radians).
    degrees : `bool`, optional
        If lonlat is True then this sets if the units are degrees or
        radians.

    Returns
    -------
    vec : `np.ndarray` (N, 3) or (3,)
        If a, b are vectors, returns a 2D array with dimensions (N, 3) with
        one vector per row.  If not vectors, a 1D array with dimensions (3,).
    """
    if lonlat:
        theta, phi = lonlat_to_thetaphi(a, b, degrees=degrees)
    else:
        _check_theta_phi(a, b)
        theta = a
        phi = b

    sin_theta = np.sin(theta)
    return np.array([sin_theta * np.cos(phi), sin_theta * np.sin(phi), np.cos(theta)]).transpose()


def vector_to_angle(vec, lonlat=True, degrees=True):
    """Convert cartesian (x, y, z) vectors to angles.

    Parameters
    ----------
    vec : `np.ndarray` (3,) or (N, 3)
        The vectors to convert to angles.
    lonlat : `bool`, optional
        Use longitude/latitude for a, b instead of co-latitude/longitude.
    degrees : `bool`, optional
        If lonlat=True then this sets if the units are degrees or radians.

    Returns
    -------
    a, b : `float` or `np.ndarray` (N,)
        Longitude/latitude (if lonlat=True) or Co-latitude(theta)/longitude(phi)
        (if lonlat=False). Longitude/latitude will be in degrees if degrees=True
        and in radians if degrees=False. Theta/phi are always in radians.
    """
    vec = np.atleast_1d(vec).reshape(-1, 3)
    norm = np.sqrt(np.sum(np.square(vec), axis=1))
    theta = np.arccos(vec[:, 2]/norm)
    phi = np.arctan2(vec[:, 1], vec[:, 0]) % (2.*np.pi)
    if lonlat:
        return thetaphi_to_lonlat(theta, phi, degrees=degrees)
    else:
        return theta, phi


def reorder(map_in, ring_to_nest=True):
    """Reorder the pixels in a map from ring to nest or nest to ring ordering.

    Parameters
    ----------
    map_in : `np.ndarray`
        Input map to reorder.
    ring_to_nest : `bool`, optional
        If True, convert ring ordering to nest ordering. If False, convert nest
        ordering to ring ordering.

    Returns
    -------
    map_out : `np.ndarray`
        Reordered map.
    """
    from ._hpgeom import ring_to_nest as convert_ring_to_nest
    from ._hpgeom import nest_to_ring as convert_nest_to_ring

    # Find the nside
    npix = map_in.size
    nside = npixel_to_nside(npix)
    # Confirm that the nside is legal for nest ordering
    _ = nside_to_order(nside)
    if (nside > 128):
        groupsize = npix // 24
    else:
        groupsize = npix

    map_out = np.zeros_like(map_in)
    for group in range(npix // groupsize):
        pixels = np.arange(group*groupsize, (group + 1)*groupsize)
        if ring_to_nest:
            map_out[convert_ring_to_nest(nside, pixels)] = map_in[pixels]
        else:
            map_out[convert_nest_to_ring(nside, pixels)] = map_in[pixels]

    return map_out



import numpy as np
import pytest

try:
    import healpy as hp
    has_healpy = True
except ImportError:
    has_healpy = False

import hpgeom


@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10])
def test_query_circle_fullsky_nest(nside):
    """Test query_circle over the full sky, nest ordering."""
    pixels = hpgeom.query_circle(nside, 0.0, 0.0, 190.0)

    np.testing.assert_array_equal(pixels, np.arange(12*nside*nside))


@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10])
def test_query_circle_fullsky_ring(nside):
    """Test query_circle over the full sky, ring ordering."""
    pixels = hpgeom.query_circle(nside, 0.0, 0.0, 190.0)

    np.testing.assert_array_equal(pixels, np.arange(12*nside*nside))


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside_radius", [(2**5, 2.0),
                                          (2**10, 1.0),
                                          (2**20, 0.01)])
@pytest.mark.parametrize("lon", [0.0, 90.0, 180.0, 270.0])
@pytest.mark.parametrize("lat", [-45.0, 0.0, 45.0, 90.0, -90.0])
def test_query_circle_nest(nside_radius, lon, lat):
    """Test query_circle, nest ordering."""
    nside = nside_radius[0]
    radius = nside_radius[1]

    # First, non-inclusive
    pixels_hpgeom = hpgeom.query_circle(nside, lon, lat, radius)

    vec = hp.ang2vec(lon, lat, lonlat=True)
    pixels_healpy = hp.query_disc(nside, vec, np.deg2rad(radius), nest=True)

    np.testing.assert_array_equal(pixels_hpgeom, pixels_healpy)

    # Second, inclusive.
    pixels_hpgeom = hpgeom.query_circle(nside, lon, lat, radius, inclusive=True)

    vec = hp.ang2vec(lon, lat, lonlat=True)
    pixels_healpy = hp.query_disc(nside, vec, np.deg2rad(radius), inclusive=True, nest=True)

    np.testing.assert_array_equal(pixels_hpgeom, pixels_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside_radius", [(2**5, 2.0),
                                          (2**10, 1.0),
                                          (2**20, 0.01),
                                          (2**10 - 2, 0.5)])
@pytest.mark.parametrize("lon", [0.0, 90.0, 180.0, 270.0])
@pytest.mark.parametrize("lat", [-45.0, 0.0, 45.0, 90.0, -90.0])
def test_query_circle_ring(nside_radius, lon, lat):
    """Test query_circle, ring ordering."""
    nside = nside_radius[0]
    radius = nside_radius[1]

    # First, non-inclusive
    pixels_hpgeom = hpgeom.query_circle(nside, lon, lat, radius, nest=False)

    vec = hp.ang2vec(lon, lat, lonlat=True)
    pixels_healpy = hp.query_disc(nside, vec, np.deg2rad(radius))

    np.testing.assert_array_equal(pixels_hpgeom, pixels_healpy)

    # Second, inclusive
    pixels_hpgeom = hpgeom.query_circle(nside, lon, lat, radius, inclusive=True, nest=False)

    vec = hp.ang2vec(lon, lat, lonlat=True)
    pixels_healpy = hp.query_disc(nside, vec, np.deg2rad(radius), inclusive=True)

    np.testing.assert_array_equal(pixels_hpgeom, pixels_healpy)


def test_query_circle_radians():
    """Test query_circle, use lonlat and radians."""
    lon = 10.0
    lat = 20.0
    radius = 1.0
    nside = 2048

    pixels_deg = hpgeom.query_circle(nside, lon, lat, radius)
    pixels_rad = hpgeom.query_circle(
        nside,
        np.deg2rad(lon),
        np.deg2rad(lat),
        np.deg2rad(radius),
        degrees=False
    )

    np.testing.assert_array_equal(pixels_rad, pixels_deg)


def test_query_circle_thetaphi():
    """Test query_circle, use theta_phi."""
    lon = 10.0
    lat = 20.0
    radius = 1.0
    nside = 2048

    pixels_deg = hpgeom.query_circle(nside, lon, lat, radius)

    theta, phi = hpgeom.lonlat_to_thetaphi(lon, lat)
    pixels_thetaphi = hpgeom.query_circle(nside, theta, phi, np.deg2rad(radius), lonlat=False)

    np.testing.assert_array_equal(pixels_thetaphi, pixels_deg)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("fact", [1, 2, 4, 8])
def test_query_circle_fact(fact):
    """Test query_circle, use other fact values."""
    lon = 10.0
    lat = 20.0
    radius = 0.5
    nside = 4096

    pixels_hpgeom = hpgeom.query_circle(nside, lon, lat, radius, inclusive=True, fact=fact)
    vec = hp.ang2vec(lon, lat, lonlat=True)
    pixels_healpy = hp.query_disc(nside, vec, np.deg2rad(radius), inclusive=True, fact=fact, nest=True)

    np.testing.assert_array_equal(pixels_hpgeom, pixels_healpy)


def test_query_circle_vec():
    """Test query_circle_vec."""
    lon = 10.0
    lat = 20.0
    radius = 0.5
    nside = 4096

    pixels_qc = hpgeom.query_circle(nside, lon, lat, radius)

    # Convert to vec
    # FIXME: use internal when available
    theta, phi = hpgeom.lonlat_to_thetaphi(lon, lat)
    sintheta = np.sin(theta)
    vec = [sintheta*np.cos(phi), sintheta*np.sin(phi), np.cos(theta)]

    pixels_qcv = hpgeom.query_circle_vec(nside, vec, np.deg2rad(radius))

    np.testing.assert_array_equal(pixels_qcv, pixels_qc)


def test_query_circle_badinputs():
    """Test query circle with bad inputs."""
    with pytest.raises(ValueError, match=r"lat .* out of range"):
        # Latitude out of range
        hpgeom.query_circle(2048, 0.0, 100.0, 1.0)

    with pytest.raises(ValueError, match=r"lat .* out of range"):
        # Latitude out of range
        hpgeom.query_circle(2048, 0.0, -100.0, 1.0)

    with pytest.raises(ValueError, match=r"colatitude \(theta\) .* out of range"):
        # theta out of range
        hpgeom.query_circle(2048, -0.1, 0.0, 0.1, lonlat=False)

    with pytest.raises(ValueError, match=r"colatitude \(theta\) .* out of range"):
        # theta out of range
        hpgeom.query_circle(2048, np.pi + 0.1, 0.0, 0.1, lonlat=False)

    with pytest.raises(ValueError, match=r"longitude \(phi\) .* out of range"):
        # phi out of range
        hpgeom.query_circle(2048, 0.0, -0.1, 0.1, lonlat=False)

    with pytest.raises(ValueError, match=r"longitude \(phi\) .* out of range"):
        # phi out of range
        hpgeom.query_circle(2048, 0.0, 2*np.pi + 0.1, 0.1, lonlat=False)

    with pytest.raises(ValueError, match=r"Radius must be positive"):
        # Radius out of range
        hpgeom.query_circle(2048, 0.0, 0.0, 0.0)

    with pytest.raises(ValueError, match=r"Inclusive factor .* must be positive"):
        # Illegal fact (must be positive)
        hpgeom.query_circle(2048, 0.0, 0.0, 1.0, inclusive=True, fact=0)

    with pytest.raises(ValueError, match=r"Inclusive factor \* nside must be \<\="):
        # Illegal fact (too large)
        hpgeom.query_circle(2**28, 0.0, 0.0, 0.0001, inclusive=True, fact=4)

    with pytest.raises(ValueError, match=r"Inclusive factor .* must be power of 2 for nest"):
        # Illegal fact (must be power of 2 for nest)
        hpgeom.query_circle(2048, 0.0, 0.0, 1.0, inclusive=True, fact=3)

    # Different platforms have different strings here, but they all say ``integer``.
    with pytest.raises(TypeError, match=r"integer"):
        # Illegal fact (must be integer)
        hpgeom.query_circle(2048, 0.0, 0.0, 1.0, inclusive=True, nest=False, fact=3.5)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
# sys.path.insert(0, os.path.abspath('..'))


# -- Project information -----------------------------------------------------

project = 'hpgeom'
copyright = '2022, Eli Rykoff (LSST Dark Energy Science Collaboration)'
author = 'Eli Rykoff'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.linkcode',
              'sphinx.ext.napoleon',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# -- Adding the API reference ------------------------------------------------

import inspect
from os.path import relpath, dirname

import hpgeom # for the relpath below

def linkcode_resolve(domain, info):
    """
    Determine the URL corresponding to Python object
    """
    if domain != 'py':
        return None

    modname = info['module']
    fullname = info['fullname']

    submod = sys.modules.get(modname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split('.'):
        try:
            obj = getattr(obj, part)
        except:
            return None

    try:
        fn = inspect.getsourcefile(obj)
    except:
        fn = None
    if not fn:
        try:
            fn = inspect.getsourcefile(sys.modules[obj.__module__])
        except:
            fn = None
    if not fn:
        return None

    try:
        source, lineno = inspect.findsource(obj)
    except:
        lineno = None

    if lineno:
        linespec = "#L%d" % (lineno + 1)
    else:
        linespec = ""

    fn = relpath(fn, start=dirname(hpgeom.__file__))

    # Could use version,release declared above here but for now we
    # just link to the latest code on the master branch.
    github = 'https://github.com/lsstdesc/hpgeom'
    return '%s/blob/main/hpgeom/%s%s' % (github, fn, linespec)



import numpy as np
import pytest

try:
    import healpy as hp
    has_healpy = True
except ImportError:
    has_healpy = False

import hpgeom


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10, 2**20, 2**29])
@pytest.mark.parametrize("pixfrac", [0.0, 0.25, 0.5, 0.75])
@pytest.mark.parametrize("step", [1, 4, 8])
@pytest.mark.parametrize("scheme", ["nest", "ring"])
def test_boundaries_single(nside, pixfrac, step, scheme):
    """Test boundaries, single pixel."""
    pix = int(pixfrac*hpgeom.nside_to_npixel(nside))
    if (scheme == "nest"):
        nest = True
    else:
        nest = False

    theta_hpgeom, phi_hpgeom = hpgeom.boundaries(nside, pix, step=step, nest=nest, lonlat=False)
    theta_healpy, phi_healpy = hp.vec2ang(hp.boundaries(nside, pix, step=step, nest=nest).T)

    assert theta_hpgeom.shape == (step*4,)
    assert phi_hpgeom.shape == (step*4,)

    np.testing.assert_array_almost_equal(theta_hpgeom, theta_healpy)
    np.testing.assert_array_almost_equal(phi_hpgeom, phi_healpy)

    lon_hpgeom, lat_hpgeom = hpgeom.boundaries(nside, pix, step=step, nest=nest, lonlat=True, degrees=True)
    lon_healpy, lat_healpy = hp.vec2ang(hp.boundaries(nside, pix, step=step, nest=nest).T, lonlat=True)

    np.testing.assert_array_almost_equal(lon_hpgeom, lon_healpy)
    np.testing.assert_array_almost_equal(lat_hpgeom, lat_healpy)

    lonrad_hpgeom, latrad_hpgeom = hpgeom.boundaries(
        nside,
        pix,
        step=step,
        nest=nest,
        lonlat=True,
        degrees=False
    )
    np.testing.assert_array_almost_equal(lonrad_hpgeom, np.deg2rad(lon_healpy))
    np.testing.assert_array_almost_equal(latrad_hpgeom, np.deg2rad(lat_healpy))


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside", [2**10, 2**20])
@pytest.mark.parametrize("pixfrac", [0.25, 0.5])
@pytest.mark.parametrize("step", [1, 4, 8])
@pytest.mark.parametrize("scheme", ["nest", "ring"])
@pytest.mark.parametrize("npix", [1, 5])
def test_boundaries_multiple(nside, pixfrac, step, scheme, npix):
    """Test boundaries, single pixel."""
    pixels = np.array([int(pixfrac*hpgeom.nside_to_npixel(nside))]*npix)
    if (scheme == "nest"):
        nest = True
    else:
        nest = False

    theta_hpgeom, phi_hpgeom = hpgeom.boundaries(nside, pixels, step=step, nest=nest, lonlat=False)
    theta_healpy, phi_healpy = hp.vec2ang(hp.boundaries(
        nside,
        pixels,
        step=step,
        nest=nest
    ).transpose([0, 2, 1]))
    theta_healpy = theta_healpy.reshape((npix, step*4))
    phi_healpy = phi_healpy.reshape((npix, step*4))

    assert theta_hpgeom.shape == (npix, step*4,)
    assert phi_hpgeom.shape == (npix, step*4,)

    np.testing.assert_array_almost_equal(theta_hpgeom, theta_healpy)
    np.testing.assert_array_almost_equal(phi_hpgeom, phi_healpy)

    lon_hpgeom, lat_hpgeom = hpgeom.boundaries(nside, pixels, step=step, nest=nest, lonlat=True, degrees=True)
    lon_healpy, lat_healpy = hp.vec2ang(hp.boundaries(
        nside,
        pixels,
        step=step,
        nest=nest,
    ).transpose([0, 2, 1]), lonlat=True)
    lon_healpy = lon_healpy.reshape((npix, step*4))
    lat_healpy = lat_healpy.reshape((npix, step*4))

    np.testing.assert_array_almost_equal(lon_hpgeom, lon_healpy)
    np.testing.assert_array_almost_equal(lat_hpgeom, lat_healpy)

    lonrad_hpgeom, latrad_hpgeom = hpgeom.boundaries(
        nside,
        pixels,
        step=step,
        nest=nest,
        lonlat=True,
        degrees=False
    )
    np.testing.assert_array_almost_equal(lonrad_hpgeom, np.deg2rad(lon_healpy))
    np.testing.assert_array_almost_equal(latrad_hpgeom, np.deg2rad(lat_healpy))


def test_boundaries_bad_inputs():
    """Test boundaries with bad inputs."""

    with pytest.raises(ValueError, match=r"nside .* must be positive"):
        hpgeom.boundaries(-1, 0)

    with pytest.raises(ValueError, match=r"Pixel value .* out of range"):
        hpgeom.boundaries(2**10, -1)

    with pytest.raises(ValueError, match=r"nside .* must be power of 2"):
        hpgeom.boundaries(2**10 - 2, -1, nest=True)



import numpy as np
import pytest

try:
    import healpy as hp
    has_healpy = True
except ImportError:
    has_healpy = False

import hpgeom


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside_delta", [(2**5, 2.0),
                                         (2**10, 1.0),
                                         (2**18, 0.01)])
@pytest.mark.parametrize("shape", ["triangle", "square", "hexagon"])
@pytest.mark.parametrize("lon_ref", [0.0, 90.0, 180.0, 270.0])
@pytest.mark.parametrize("lat_ref", [-43.0, 0.0, 43.0, 85.0, -85.0])
def test_query_polygon_nest(nside_delta, shape, lon_ref, lat_ref):
    """Test query_polygon, nest ordering."""
    nside = nside_delta[0]
    delta = nside_delta[1]

    if shape == "triangle":
        lon = np.array([lon_ref, lon_ref + delta, lon_ref + delta])
        lat = np.array([lat_ref, lat_ref + delta, lat_ref - delta])
    elif shape == "square":
        lon = np.array([lon_ref, lon_ref + delta, lon_ref + delta, lon_ref])
        lat = np.array([lat_ref, lat_ref, lat_ref + delta, lat_ref + delta])
    else:
        lon = np.array([
            lon_ref,
            lon_ref + delta/2.,
            lon_ref + delta,
            lon_ref + 3*delta/2.,
            lon_ref + delta,
            lon_ref + delta/2.,
        ])
        lat = np.array([
            lat_ref,
            lat_ref + delta,
            lat_ref + delta,
            lat_ref,
            lat_ref - delta,
            lat_ref - delta
        ])

    # Test forward, non-inclusive
    pixels_hpgeom = hpgeom.query_polygon(nside, lon, lat)
    vec = hpgeom.angle_to_vector(lon, lat, lonlat=True)
    pixels_healpy = hp.query_polygon(nside, vec, nest=True)

    np.testing.assert_array_equal(pixels_hpgeom, pixels_healpy)

    # Test forward, non-inclusive, closed polygon
    lon2 = np.append(lon, lon[0])
    lat2 = np.append(lat, lat[0])

    pixels_hpgeom2 = hpgeom.query_polygon(nside, lon2, lat2)
    # Note healpy does not support closed polygons, so we use the first
    # run as a comparison.

    np.testing.assert_array_equal(pixels_hpgeom2, pixels_hpgeom)

    # Test reversed, non-inclusive
    pixels_hpgeom = hpgeom.query_polygon(nside, lon[::-1], lat[::-1])
    vec = hpgeom.angle_to_vector(lon[::-1], lat[::-1], lonlat=True)
    pixels_healpy = hp.query_polygon(nside, vec, nest=True)

    np.testing.assert_array_equal(pixels_hpgeom, pixels_healpy)

    # Test forward, inclusive
    pixels_hpgeom = hpgeom.query_polygon(nside, lon, lat, inclusive=True)
    vec = hpgeom.angle_to_vector(lon, lat, lonlat=True)
    pixels_healpy = hp.query_polygon(nside, vec, nest=True, inclusive=True)

    np.testing.assert_array_equal(pixels_hpgeom, pixels_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside_delta", [
    (2**5, 2.0),
    (2**10, 1.0),
    (2**18, 0.01),
    (2**10 - 2, 1.0)
])
@pytest.mark.parametrize("shape", ["triangle", "square", "hexagon"])
@pytest.mark.parametrize("lon_ref", [0.0, 90.0, 180.0])
@pytest.mark.parametrize("lat_ref", [-44.0, 0.0, 44.0, 85.0, -85.0])
def test_query_polygon_ring(nside_delta, shape, lon_ref, lat_ref):
    """Test query_polygon, ring ordering."""
    nside = nside_delta[0]
    delta = nside_delta[1]

    if shape == "triangle":
        lon = np.array([lon_ref, lon_ref + delta, lon_ref + delta])
        lat = np.array([lat_ref, lat_ref + delta, lat_ref - delta])
    elif shape == "square":
        lon = np.array([lon_ref, lon_ref + delta, lon_ref + delta, lon_ref])
        lat = np.array([lat_ref, lat_ref, lat_ref + delta, lat_ref + delta])
    else:
        lon = np.array([
            lon_ref,
            lon_ref + delta/2.,
            lon_ref + delta,
            lon_ref + 3*delta/2.,
            lon_ref + delta,
            lon_ref + delta/2.,
        ])
        lat = np.array([
            lat_ref,
            lat_ref + delta,
            lat_ref + delta,
            lat_ref,
            lat_ref - delta,
            lat_ref - delta
        ])

    # Test forward, non-inclusive
    pixels_hpgeom = hpgeom.query_polygon(nside, lon, lat, nest=False)
    vec = hpgeom.angle_to_vector(lon, lat, lonlat=True)
    pixels_healpy = hp.query_polygon(nside, vec, nest=False)

    np.testing.assert_array_equal(pixels_hpgeom, pixels_healpy)

    # Test reversed, non-inclusive
    pixels_hpgeom_rev = hpgeom.query_polygon(nside, lon[::-1], lat[::-1], nest=False)
    vec = hpgeom.angle_to_vector(lon[::-1], lat[::-1], lonlat=True)
    pixels_healpy = hp.query_polygon(nside, vec, nest=False)

    np.testing.assert_array_equal(pixels_hpgeom_rev, pixels_healpy)

    # Test forward, inclusive
    pixels_hpgeom = hpgeom.query_polygon(nside, lon, lat, inclusive=True, nest=False)
    vec = hpgeom.angle_to_vector(lon, lat, lonlat=True)
    pixels_healpy = hp.query_polygon(nside, vec, nest=False, inclusive=True)

    np.testing.assert_array_equal(pixels_hpgeom, pixels_healpy)


def test_query_polygon_radians():
    """Test query_polygon, use lonlat and radians."""
    nside = 1024
    delta = 1.0
    lon_ref = 180.0
    lat_ref = 0.0
    lon = np.array([lon_ref, lon_ref + delta, lon_ref + delta, lon_ref])
    lat = np.array([lat_ref, lat_ref, lat_ref + delta, lat_ref + delta])

    pixels_deg = hpgeom.query_polygon(nside, lon, lat)
    pixels_rad = hpgeom.query_polygon(nside, np.deg2rad(lon), np.deg2rad(lat), degrees=False)

    np.testing.assert_array_equal(pixels_deg, pixels_rad)


def test_query_polygon_thetaphi():
    """Test query_polygon, use theta_phi."""
    nside = 1024
    delta = 1.0
    lon_ref = 180.0
    lat_ref = 0.0
    lon = np.array([lon_ref, lon_ref + delta, lon_ref + delta, lon_ref])
    lat = np.array([lat_ref, lat_ref, lat_ref + delta, lat_ref + delta])

    pixels_lonlat = hpgeom.query_polygon(nside, lon, lat)
    theta, phi = hpgeom.lonlat_to_thetaphi(lon, lat)
    pixels_thetaphi = hpgeom.query_polygon(nside, theta, phi, lonlat=False)

    np.testing.assert_array_equal(pixels_lonlat, pixels_thetaphi)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("fact", [1, 2, 4, 8])
def test_query_polygon_fact(fact):
    """Test query_polygon, use other fact values."""
    nside = 1024
    delta = 1.0
    lon_ref = 180.0
    lat_ref = 0.0
    lon = np.array([lon_ref, lon_ref + delta, lon_ref + delta, lon_ref])
    lat = np.array([lat_ref, lat_ref, lat_ref + delta, lat_ref + delta])

    pixels_hpgeom = hpgeom.query_polygon(nside, lon, lat, fact=fact)
    vec = hpgeom.angle_to_vector(lon, lat)
    pixels_healpy = hp.query_polygon(nside, vec, nest=True, fact=fact)

    np.testing.assert_array_equal(pixels_hpgeom, pixels_healpy)


def test_query_polygon_vec():
    """Test query_polygon_vec."""
    nside = 1024
    delta = 1.0
    lon_ref = 180.0
    lat_ref = 0.0
    lon = np.array([lon_ref, lon_ref + delta, lon_ref + delta, lon_ref])
    lat = np.array([lat_ref, lat_ref, lat_ref + delta, lat_ref + delta])

    pixels_lonlat = hpgeom.query_polygon(nside, lon, lat)
    vec = hpgeom.angle_to_vector(lon, lat)
    pixels_vec = hpgeom.query_polygon_vec(nside, vec)

    np.testing.assert_array_equal(pixels_lonlat, pixels_vec)


def test_query_polygon_badinputs():
    """Test query_polygon with bad inputs."""
    nside = 1024
    delta = 1.0
    lon_ref = 180.0
    lat_ref = 0.0
    lon = np.array([lon_ref, lon_ref + delta, lon_ref + delta, lon_ref])
    lat = np.array([lat_ref, lat_ref, lat_ref + delta, lat_ref + delta])

    _lat = lat.copy()
    _lat[0] = 100.0
    with pytest.raises(ValueError, match=r"lat .* out of range"):
        hpgeom.query_polygon(nside, lon, _lat)

    _lat = lat.copy()
    _lat[0] = -100.0
    with pytest.raises(ValueError, match=r"lat .* out of range"):
        hpgeom.query_polygon(nside, lon, _lat)

    _theta, _phi = hpgeom.lonlat_to_thetaphi(lon, lat)
    _theta[0] = -0.1
    with pytest.raises(ValueError, match=r"colatitude \(theta\) .* out of range"):
        hpgeom.query_polygon(nside, _theta, _phi, lonlat=False)

    _lon = np.append(lon, lon_ref + delta/2.)
    _lat = np.append(lat, lat_ref + delta/2.)
    with pytest.raises(RuntimeError, match="Polygon is not convex"):
        hpgeom.query_polygon(nside, _lon, _lat)

    with pytest.raises(ValueError, match=r"Inclusive factor .* must be positive"):
        hpgeom.query_polygon(nside, lon, lat, inclusive=True, fact=0)

    with pytest.raises(ValueError, match=r"Inclusive factor \* nside must be \<\="):
        hpgeom.query_polygon(2**28, lon, lat, inclusive=True, fact=4)

    with pytest.raises(ValueError, match=r"Inclusive factor .* must be power of 2 for nest"):
        hpgeom.query_polygon(nside, lon, lat, inclusive=True, fact=3)



import numpy as np
import pytest

try:
    import healpy as hp
    has_healpy = True
except ImportError:
    has_healpy = False

import hpgeom


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10, 2**15, 2**20, 2**25, 2**29])
@pytest.mark.parametrize("scheme", ["nest", "ring"])
def test_neighbors(nside, scheme):
    """Test neighbors."""
    np.random.seed(12345)

    if scheme == "nest":
        nest = True
    else:
        nest = False

    pix = np.random.randint(low=0, high=12*nside*nside-1, size=1_000_000, dtype=np.int64)

    neighbors_hpgeom = hpgeom.neighbors(nside, pix, nest=nest)
    neighbors_healpy = hp.get_all_neighbours(nside, pix, nest=nest).T

    np.testing.assert_array_equal(neighbors_hpgeom, neighbors_healpy)


def test_neighbors_bad_input():
    """Test neighbors, bad input."""
    with pytest.raises(ValueError, match=r"Pixel value .* out of range"):
        hpgeom.neighbors(1024, -1)

    with pytest.raises(ValueError, match=r"nside .* must be positive"):
        hpgeom.neighbors(-1, 100)



import numpy as np
import pytest
import warnings

import hpgeom

from utils import match_arrays


EPSILON = 2.220446049250313e-16


def _pos_in_box(lon, lat, lon0, lon1, lat0, lat1):
    """Check which positions are in a box.

    Parameters
    ----------
    lon, lat : `np.ndarray`
    lon0, lon1, lat0, lat1 : `float`

    Returns
    -------
    indices : `np.ndarray`
    """
    _lon0 = (np.deg2rad(lon0) - EPSILON) % (2.*np.pi)
    _lon1 = np.deg2rad(lon1) % (2.*np.pi)
    _lat0 = np.deg2rad(lat0) - EPSILON
    _lat1 = np.deg2rad(lat1)

    _lon = np.deg2rad(lon)
    _lat = np.deg2rad(lat)

    if _lon0 < _lon1:
        inside = ((_lon >= _lon0) & (_lon < _lon1) & (_lat >= _lat0) & (_lat < _lat1))
    else:
        inside = (((_lon < _lon1) | (_lon >= _lon0)) & (_lat >= _lat0) & (_lat < _lat1))

    return inside


@pytest.mark.parametrize("nside_radius", [(2**7, 2.0),
                                          (2**10, 1.0),
                                          (2**20, 0.01)])
@pytest.mark.parametrize("lon", [0.0, 90.0, 180.0, 270.0])
@pytest.mark.parametrize("lat", [-45.0, 0.0, 45.0])
def test_query_box_square(nside_radius, lon, lat):
    """Test query_box with squares."""
    nside = nside_radius[0]
    radius = nside_radius[1]

    box = [lon - radius, lon + radius, lat - radius, lat + radius]

    pixels = hpgeom.query_box(nside, *box)

    pixels_circle = hpgeom.query_circle(nside, lon, lat, radius*2.5)
    lon_circle, lat_circle = hpgeom.pixel_to_angle(nside, pixels_circle)
    inside = _pos_in_box(lon_circle, lat_circle, *box)

    np.testing.assert_array_equal(pixels, pixels_circle[inside])

    # Do inclusive
    pixels_box = hpgeom.query_box(nside, *box, inclusive=True)

    # Ensure all the inner pixels are in the inclusive pixels
    sub1, sub2 = match_arrays(pixels_box, pixels)
    assert sub2.size == pixels.size

    # Look at boundaries of the pixels, check if any are included.
    pixels_circle = hpgeom.query_circle(nside, lon, lat, radius*2.5)
    boundaries_lon, boundaries_lat = hpgeom.boundaries(nside, pixels_circle, step=4)
    cut = _pos_in_box(boundaries_lon.ravel(), boundaries_lat.ravel(), *box)
    test = cut.reshape(boundaries_lon.shape).sum(axis=1)
    pixels_circle_box = pixels_circle[test > 0]

    # Ensure all these pixels are in the inclusive list
    sub1, sub2 = match_arrays(pixels_circle_box, pixels_box)
    assert sub1.size == pixels_circle_box.size


@pytest.mark.parametrize("nside_radius", [(2**7, 2.0),
                                          (2**10, 1.0),
                                          (2**20, 0.01)])
@pytest.mark.parametrize("lon", [0.0, 360.0])
@pytest.mark.parametrize("lat", [-45.0, 0.0, 45.0])
def test_query_box_edge(nside_radius, lon, lat):
    """Test query_box at the longitude edges."""
    nside = nside_radius[0]
    radius = nside_radius[1]

    if (lon == 0.0):
        box = [lon, lon + 2*radius, lat - radius, lat + radius]
    else:
        box = [lon - 2*radius, lon, lat - radius, lat + radius]

    pixels = hpgeom.query_box(nside, *box)

    pixels_circle = hpgeom.query_circle(nside, (box[0] + box[1])/2., lat, radius*2.5)
    lon_circle, lat_circle = hpgeom.pixel_to_angle(nside, pixels_circle)
    inside = _pos_in_box(lon_circle, lat_circle, *box)

    np.testing.assert_array_equal(pixels, pixels_circle[inside])

    # Do inclusive
    pixels_box = hpgeom.query_box(nside, *box, inclusive=True)

    # Ensure all the inner pixels are in the inclusive pixels
    sub1, sub2 = match_arrays(pixels_box, pixels)
    assert sub2.size == pixels.size

    # Look at boundaries of the pixels.
    pixels_circle = hpgeom.query_circle(nside, (box[0] + box[1])/2., lat, radius*2.5)
    boundaries_lon, boundaries_lat = hpgeom.boundaries(nside, pixels_circle, step=4)
    cut = _pos_in_box(boundaries_lon.ravel(), boundaries_lat.ravel(), *box)
    test = cut.reshape(boundaries_lon.shape).sum(axis=1)
    pixels_circle_box = pixels_circle[test > 0]

    # Ensure all these pixels are in the inclusive list
    sub1, sub2 = match_arrays(pixels_circle_box, pixels_box)
    assert sub1.size == pixels_circle_box.size

    # And ensure that we did wrap over the edge
    lon_box, lat_box = hpgeom.pixel_to_angle(nside, pixels_box)
    if (lon == 0.0):
        assert lon_box.max() > 355.0
    else:
        assert lon_box.min() < 5.0


@pytest.mark.parametrize("nside_radius", [(2**5, 2.0),
                                          (2**10, 1.0),
                                          (2**20, 0.01)])
@pytest.mark.parametrize("lon", [0.0, 90.0, 180.0, 270.0])
@pytest.mark.parametrize("lat", [90.0, -90.0])
def test_query_box_pole(nside_radius, lon, lat):
    """Test query_box at the latitude poles."""
    nside = nside_radius[0]
    radius = nside_radius[1]

    if (lat == 90.0):
        box = [lon - radius, lon + radius, lat - 2*radius, lat]
    else:
        box = [lon - radius, lon + radius, lat, lat + 2*radius]

    pixels = hpgeom.query_box(nside, *box)

    pixels_circle = hpgeom.query_circle(nside, lon, (box[2] + box[3])/2., radius*2.5)
    lon_circle, lat_circle = hpgeom.pixel_to_angle(nside, pixels_circle)
    inside = _pos_in_box(lon_circle, lat_circle, *box)

    np.testing.assert_array_equal(pixels, pixels_circle[inside])

    # Do inclusive
    pixels_box = hpgeom.query_box(nside, *box, inclusive=True)

    # Ensure all the inner pixels are in the inclusive pixels
    sub1, sub2 = match_arrays(pixels_box, pixels)
    assert sub2.size == pixels.size

    # Look at boundaries of the pixels, check if any are included.
    pixels_circle = hpgeom.query_circle(nside, lon, lat, radius*2.5)
    boundaries_lon, boundaries_lat = hpgeom.boundaries(nside, pixels_circle, step=4)
    cut = _pos_in_box(boundaries_lon.ravel(), boundaries_lat.ravel(), *box)
    test = cut.reshape(boundaries_lon.shape).sum(axis=1)
    pixels_circle_box = pixels_circle[test > 0]

    # Ensure all these pixels are in the inclusive list
    sub1, sub2 = match_arrays(pixels_circle_box, pixels_box)
    if (lon == 0.0 and lat == -90.0):
        # There is some oddity with the polar pixel boundaries that needs to
        # be investigated.
        assert sub1.size == (pixels_circle_box.size - 2)
    else:
        assert sub1.size == pixels_circle_box.size


@pytest.mark.parametrize("nside", [2**5, 2**10])
@pytest.mark.parametrize("lat", [90.0, -90.0])
def test_query_box_pole_full_longitude(nside, lat):
    """Test query_box at the latitude poles, full longitude."""
    radius = 2.0

    if (lat == 90.0):
        box = [0.0, 360.0, lat - radius, lat]
    else:
        box = [0.0, 360.0, lat, lat + radius]

    pixels = hpgeom.query_box(nside, *box)

    # This should be the same as a circle at the pole.
    pixels_circle = hpgeom.query_circle(nside, 0.0, lat, radius)

    np.testing.assert_array_equal(pixels, pixels_circle)

    # Do inclusive
    pixels_box = hpgeom.query_box(nside, *box, inclusive=True)

    # Ensure all the inner pixels are in the inclusive pixels
    sub1, sub2 = match_arrays(pixels_box, pixels)
    assert sub2.size == pixels.size

    pixels_circle_inc = hpgeom.query_circle(nside, 0.0, lat, radius, inclusive=True)

    np.testing.assert_array_equal(pixels_box, pixels_circle_inc)


@pytest.mark.parametrize("nside", [2**5, 2**10])
@pytest.mark.parametrize("lat", [-45.0, 0.0, 45.0])
def test_query_box_full_longitude(nside, lat):
    """Test query_box at the latitude poles."""
    radius = 1.0

    box = [0.0, 360.0, lat - radius, lat + radius]

    pixels = hpgeom.query_box(nside, *box)

    pixels_all = np.arange(hpgeom.nside_to_npixel(nside))
    lon_all, lat_all = hpgeom.pixel_to_angle(nside, pixels_all)
    inside = ((lat_all >= box[2]) & (lat_all < box[3]))

    np.testing.assert_array_equal(pixels, pixels_all[inside])

    # Do inclusive
    pixels_box = hpgeom.query_box(nside, *box, inclusive=True)

    # Ensure all the inner pixels are in the inclusive pixels
    sub1, sub2 = match_arrays(pixels_box, pixels)
    assert sub2.size == pixels.size

    # Look at boundaries of the pixels, check if any are included.
    inside_plus = ((lat_all >= (lat - 2*radius)) & (lat_all <= (lat + 2*radius)))
    boundaries_lon, boundaries_lat = hpgeom.boundaries(nside, pixels_all[inside_plus], step=4)
    cut = ((boundaries_lat.ravel() >= box[2]) & (boundaries_lat.ravel() < box[3]))
    test = cut.reshape(boundaries_lat.shape).sum(axis=1)
    pixels_all_box = pixels_all[inside_plus][test > 0]

    # Ensure all these pixels are in the inclusive list
    sub1, sub2 = match_arrays(pixels_all_box, pixels_box)
    assert sub1.size == pixels_all_box.size


@pytest.mark.parametrize("fact", [1, 2, 4, 8])
def test_query_box_fact(fact):
    """Test query_box, use other fact values."""
    nside = 1024
    radius = 1.0
    lon = 90.0
    lat = 0.0

    box = [lon - radius, lon + radius, lat - radius, lat + radius]

    pixels = hpgeom.query_box(nside, *box)

    pixels_box = hpgeom.query_box(nside, *box, inclusive=True, fact=fact)

    # Ensure all the inner pixels are in the inclusive pixels
    sub1, sub2 = match_arrays(pixels_box, pixels)
    assert sub2.size == pixels.size

    # Look at boundaries of the pixels, check if any are included.
    pixels_circle = hpgeom.query_circle(nside, lon, lat, radius*2.5)
    boundaries_lon, boundaries_lat = hpgeom.boundaries(nside, pixels_circle, step=4)
    cut = _pos_in_box(boundaries_lon.ravel(), boundaries_lat.ravel(), *box)
    test = cut.reshape(boundaries_lon.shape).sum(axis=1)
    pixels_circle_box = pixels_circle[test > 0]

    # Ensure all these pixels are in the inclusive list
    sub1, sub2 = match_arrays(pixels_circle_box, pixels_box)
    assert sub1.size == pixels_circle_box.size


def test_query_box_empty():
    """Test query box when it is empty."""

    pixels = hpgeom.query_box(1024, 25.0, 25.0, 0.0, 2.0)
    assert pixels.size == 0

    pixels = hpgeom.query_box(1024, 90.0, 92.0, 45.0, 45.0)
    assert pixels.size == 0


def test_query_box_radians():
    """Test query_box, use lonlat and radians."""
    nside = 1024
    lon = 90.0
    lat = 45.0
    radius = 1.0

    box = [lon - radius, lon + radius, lat - radius, lat + radius]

    pixels_deg = hpgeom.query_box(nside, *box)
    pixels_rad = hpgeom.query_box(nside, *np.deg2rad(box), degrees=False)

    np.testing.assert_array_equal(pixels_rad, pixels_deg)


def test_query_box_thetaphi():
    """Test query_box, use theta_phi."""
    nside = 1024
    lon = 90.0
    lat = 45.0
    radius = 1.0

    box = [lon - radius, lon + radius, lat - radius, lat + radius]

    pixels_deg = hpgeom.query_box(nside, *box)
    theta, phi = hpgeom.lonlat_to_thetaphi(np.array(box[0: 2]), np.array(box[2:]))
    box_thetaphi = [theta[1], theta[0], phi[0], phi[1]]
    pixels_thetaphi = hpgeom.query_box(nside, *box_thetaphi, lonlat=False)

    np.testing.assert_array_equal(pixels_thetaphi, pixels_deg)


def test_query_box_ring():
    """Test query_box, ring ordering."""
    nside = 1024
    lon = 90.0
    lat = 45.0
    radius = 1.0

    box = [lon - radius, lon + radius, lat - radius, lat + radius]

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        pixels_ring = hpgeom.query_box(nside, *box, nest=False)

    pixels_nest = hpgeom.query_box(nside, *box)
    pixels_nest_to_ring = hpgeom.nest_to_ring(nside, pixels_nest)
    pixels_nest_to_ring.sort()

    np.testing.assert_array_equal(pixels_ring, pixels_nest_to_ring)


def test_query_box_badinputs():
    """Test query_box with bad inputs."""
    with pytest.raises(ValueError, match=r"lat .* out of range"):
        hpgeom.query_box(2048, 0.0, 10.0, 80.0, 100.0)

    with pytest.raises(ValueError, match=r"lat .* out of range"):
        hpgeom.query_box(2048, 0.0, 10.0, -100.0, -80.0)

    with pytest.raises(ValueError, match=r"colatitude \(theta\) .* out of range"):
        hpgeom.query_box(2048, -0.1, 0.0, 0.0, 0.2, lonlat=False)

    with pytest.raises(ValueError, match=r"colatitude \(theta\) .* out of range"):
        hpgeom.query_box(2048, 0.0, np.pi + 1, 0.0, 0.2, lonlat=False)

    with pytest.raises(ValueError, match=r"longitude \(phi\) .* out of range"):
        hpgeom.query_box(2048, 0.0, 0.1, -0.1, 0.2, lonlat=False)

    with pytest.raises(ValueError, match=r"longitude \(phi\) .* out of range"):
        hpgeom.query_box(2048, 0.0, 0.1, 0.1, 3*np.pi, lonlat=False)

    with pytest.raises(ValueError, match=r"b1/lat1 must be >= b0/lat0"):
        hpgeom.query_box(2048, 0.0, 0.1, 1.0, 0.0)

    with pytest.raises(ValueError, match=r"a1/colatitude1 must be <= a0/colatitude0"):
        hpgeom.query_box(2048, 1.0, 0.9, 0.0, 0.1, lonlat=False)

    with pytest.raises(ValueError, match=r"Inclusive factor .* must be positive"):
        hpgeom.query_box(2048, 0.0, 1.0, 0.0, 1.0, inclusive=True, fact=0)

    with pytest.raises(ValueError, match=r"Inclusive factor \* nside must be \<\="):
        hpgeom.query_box(2**28, 0.0, 1.0, 0.0, 1.0, inclusive=True, fact=4)

    with pytest.raises(ValueError, match=r"Inclusive factor .* must be power of 2 for nest"):
        hpgeom.query_box(2048, 0.0, 1.0, 0.0, 1.0, inclusive=True, fact=3)

    # Different platforms have different strings here, but they all say ``integer``.
    with pytest.raises(TypeError, match=r"integer"):
        hpgeom.query_box(2048, 0.0, 1.0, 0.0, 1.0, inclusive=True, nest=False, fact=3.5)

    # Check resource warning
    with pytest.warns(ResourceWarning, match=r"natively supports nest ordering"):
        hpgeom.query_box(1024, 0.0, 1.0, 0.0, 1.0, nest=False)



import numpy as np
import pytest

try:
    import healpy as hp
    has_healpy = True
except ImportError:
    has_healpy = False

import hpgeom


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("degrees", [False, True])
def test_max_pixel_radius(degrees):
    """Test max_pixel_radius."""
    nsides = 2**np.arange(29)
    nsides_offset = nsides + 2

    # Compute array-like
    radii_hpgeom = hpgeom.max_pixel_radius(nsides, degrees=degrees)
    radii_healpy = hp.max_pixrad(nsides, degrees=degrees)

    np.testing.assert_array_almost_equal(radii_hpgeom, radii_healpy)

    # Compute with offset, ring-like.
    # Note that we can't compare this to healpy because healpy crashes.
    radii_hpgeom = hpgeom.max_pixel_radius(nsides_offset, degrees=degrees)

    # Compute individual
    radius_hpgeom = hpgeom.max_pixel_radius(1024, degrees=degrees)
    radius_healpy = hp.max_pixrad(1024, degrees=degrees)

    np.testing.assert_almost_equal(radius_hpgeom, radius_healpy)


def test_max_pixel_radius_badinputs():
    """Test max_pixel_radius with bad inputs."""
    with pytest.raises(ValueError, match=r"nside .* must be positive"):
        hpgeom.max_pixel_radius(-10)



import numpy as np
import pytest

try:
    import healpy as hp
    has_healpy = True
except ImportError:
    has_healpy = False

import hpgeom


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_thetaphi_to_lonlat():
    """Test thetaphi_to_lonlat."""
    np.random.seed(12345)

    theta = np.random.uniform(0, np.pi, 1_000_000)
    phi = np.random.uniform(0, 2*np.pi, 1_000_000)

    lon_hpgeom, lat_hpgeom = hpgeom.thetaphi_to_lonlat(float(theta[0]), float(phi[0]))
    lon_healpy, lat_healpy = hp.pixelfunc.thetaphi2lonlat(float(theta[0]), float(phi[0]))

    np.testing.assert_almost_equal(lon_hpgeom, lon_healpy)
    np.testing.assert_almost_equal(lat_hpgeom, lat_healpy)

    lon_hpgeom, lat_hpgeom = hpgeom.thetaphi_to_lonlat(theta, phi)
    lon_healpy, lat_healpy = hp.pixelfunc.thetaphi2lonlat(theta, phi)

    np.testing.assert_array_almost_equal(lon_hpgeom, lon_healpy)
    np.testing.assert_array_almost_equal(lat_hpgeom, lat_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_lonlat_to_thetaphi():
    """Test thetaphi_to_lonlat."""
    np.random.seed(12345)

    lon = np.random.uniform(0, 360.0, 1_000_000)
    lat = np.random.uniform(-90.0, 90.0, 1_000_000)

    theta_hpgeom, phi_hpgeom = hpgeom.thetaphi_to_lonlat(float(lon[0]), float(lat[0]))
    theta_healpy, phi_healpy = hp.pixelfunc.thetaphi2lonlat(float(lon[0]), float(lat[0]))

    np.testing.assert_almost_equal(theta_hpgeom, theta_healpy)
    np.testing.assert_almost_equal(phi_hpgeom, phi_healpy)

    theta_hpgeom, phi_hpgeom = hpgeom.thetaphi_to_lonlat(lon, lat)
    theta_healpy, phi_healpy = hp.pixelfunc.thetaphi2lonlat(lon, lat)

    np.testing.assert_array_almost_equal(theta_hpgeom, theta_healpy)
    np.testing.assert_array_almost_equal(phi_hpgeom, phi_healpy)



try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:
    from importlib_metadata import version, PackageNotFoundError

try:
    __version__ = version("hpgeom")
except PackageNotFoundError:
    # package is not installed
    pass

from . import hpgeom

from .hpgeom import *



import numpy as np
import pytest

try:
    import healpy as hp
    has_healpy = True
except ImportError:
    has_healpy = False

import hpgeom


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10, 2**15, 2**20, 2**25, 2**29, 2**10 - 100])
def test_angle_to_pixel_ring(nside):
    """Test angle_to_pixel for ring scheme."""
    np.random.seed(12345)

    lon = np.random.uniform(low=0.0, high=360.0, size=1_000_000)
    lat = np.random.uniform(low=-90.0, high=90.0, size=1_000_000)

    pix_hpgeom = hpgeom.angle_to_pixel(nside, lon, lat, nest=False, lonlat=True, degrees=True)
    pix_healpy = hp.ang2pix(nside, lon, lat, nest=False, lonlat=True)

    np.testing.assert_array_equal(pix_hpgeom, pix_healpy)

    pix_hpgeom = hpgeom.angle_to_pixel(
        nside,
        np.deg2rad(lon),
        np.deg2rad(lat),
        nest=False,
        lonlat=True,
        degrees=False
    )

    np.testing.assert_array_equal(pix_hpgeom, pix_healpy)

    phi = np.deg2rad(lon)
    theta = -np.deg2rad(lat) + np.pi/2.

    pix_hpgeom = hpgeom.angle_to_pixel(nside, theta, phi, nest=False, lonlat=False)
    pix_healpy = hp.ang2pix(nside, theta, phi, nest=False, lonlat=False)

    np.testing.assert_array_equal(pix_hpgeom, pix_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10, 2**15, 2**20, 2**25, 2**29])
def test_angle_to_pixel_nest(nside):
    """Test angle_to_pixel for nest scheme."""
    np.random.seed(12345)

    lon = np.random.uniform(low=0.0, high=360.0, size=1_000_000)
    lat = np.random.uniform(low=-90.0, high=90.0, size=1_000_000)

    pix_hpgeom = hpgeom.angle_to_pixel(nside, lon, lat, nest=True, lonlat=True, degrees=True)
    pix_healpy = hp.ang2pix(nside, lon, lat, nest=True, lonlat=True)

    np.testing.assert_array_equal(pix_hpgeom, pix_healpy)

    pix_hpgeom = hpgeom.angle_to_pixel(
        nside,
        np.deg2rad(lon),
        np.deg2rad(lat),
        nest=True,
        lonlat=True,
        degrees=False
    )

    np.testing.assert_array_equal(pix_hpgeom, pix_healpy)

    theta, phi = hpgeom.lonlat_to_thetaphi(lon, lat)

    pix_hpgeom = hpgeom.angle_to_pixel(nside, theta, phi, nest=True, lonlat=False)
    pix_healpy = hp.ang2pix(nside, theta, phi, nest=True, lonlat=False)

    np.testing.assert_array_equal(pix_hpgeom, pix_healpy)


@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10, 2**15, 2**20, 2**25, 2**29])
def test_angle_to_pixel_scalar(nside):
    """Test angle_to_pixel for scalars."""
    np.random.seed(12345)

    lon = np.random.uniform(low=0.0, high=360.0, size=100)
    lat = np.random.uniform(low=-90.0, high=90.0, size=100)

    pix_arr = hpgeom.angle_to_pixel(nside, lon, lat, nest=True, lonlat=True, degrees=True)

    pix_scalar1 = hpgeom.angle_to_pixel(nside, lon[0], lat[0], nest=True, lonlat=True, degrees=True)

    assert pix_scalar1 == pix_arr[0]

    pix_scalar2 = hpgeom.angle_to_pixel(
        nside,
        float(lon[0]),
        float(lat[0]),
        nest=True,
        lonlat=True,
        degrees=True
    )

    assert pix_scalar2 == pix_arr[0]


def test_angle_to_pixel_mismatched_dims():
    """Test angle_to_pixel errors when dimensions are mismatched."""
    np.random.seed(12345)

    lon = np.random.uniform(low=0.0, high=360.0, size=100)
    lat = np.random.uniform(low=-90.0, high=90.0, size=100)

    with pytest.raises(ValueError, match=r"arrays could not be broadcast together"):
        hpgeom.angle_to_pixel([2048, 4096], lon[0], lat)

    with pytest.raises(ValueError, match=r"arrays could not be broadcast together"):
        hpgeom.angle_to_pixel([2048, 4096], lon, lat[0])

    with pytest.raises(ValueError, match=r"arrays could not be broadcast together"):
        hpgeom.angle_to_pixel(2048, lon[0: 5], lat)

    with pytest.raises(ValueError, match=r"arrays could not be broadcast together"):
        hpgeom.angle_to_pixel(2048, lon, lat[0: 5])

    with pytest.raises(ValueError, match=r"arrays could not be broadcast together"):
        hpgeom.angle_to_pixel(2048, lon.reshape((10, 10)), lat)

    with pytest.raises(ValueError, match=r"arrays could not be broadcast together"):
        hpgeom.angle_to_pixel(2048, lon, lat.reshape((10, 10)))


def test_angle_to_pixel_bad_nside():
    """Test angle_to_pixel errors when given a bad nside."""
    np.random.seed(12345)

    lon = np.random.uniform(low=0.0, high=360.0, size=100)
    lat = np.random.uniform(low=-90.0, high=90.0, size=100)

    with pytest.raises(ValueError, match=r"nside .* must be positive"):
        hpgeom.angle_to_pixel(-10, lon, lat, nest=False)

    with pytest.raises(ValueError, match=r"nside .* must be positive"):
        hpgeom.angle_to_pixel(-10, lon, lat, nest=True)

    with pytest.raises(ValueError, match=r"nside .* must be power of 2"):
        hpgeom.angle_to_pixel(2040, lon, lat, nest=True)

    with pytest.raises(ValueError, match=r"nside .* must not be greater"):
        hpgeom.angle_to_pixel(2**30, lon, lat, nest=True)


def test_angle_to_pixel_bad_coords():
    """Test angle_to_pixel errors when given bad coords."""
    with pytest.raises(ValueError, match=r"lat .* out of range"):
        # Dec out of range
        hpgeom.angle_to_pixel(2048, 0.0, 100.0)

    with pytest.raises(ValueError, match=r"lat .* out of range"):
        # Dec out of range
        hpgeom.angle_to_pixel(2048, 0.0, -100.0)

    with pytest.raises(ValueError, match=r"colatitude \(theta\) .* out of range"):
        # theta out of range
        hpgeom.angle_to_pixel(2048, -0.1, 0.0, lonlat=False)

    with pytest.raises(ValueError, match=r"colatitude \(theta\) .* out of range"):
        # theta out of range
        hpgeom.angle_to_pixel(2048, np.pi + 0.1, 0.0, lonlat=False)

    with pytest.raises(ValueError, match=r"longitude \(phi\) .* out of range"):
        # phi out of range
        hpgeom.angle_to_pixel(2048, 0.0, -0.1, lonlat=False)

    with pytest.raises(ValueError, match=r"longitude \(phi\) .* out of range"):
        # phi out of range
        hpgeom.angle_to_pixel(2048, 0.0, 2*np.pi + 0.1, lonlat=False)



import numpy as np
import pytest

try:
    import healpy as hp
    has_healpy = True
except ImportError:
    has_healpy = False

import hpgeom


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10, 2**15, 2**20, 2**25, 2**29])
@pytest.mark.parametrize("scheme", ["nest", "ring"])
def test_vector_to_pixel(nside, scheme):
    """Test vector_to_pixel."""
    np.random.seed(12345)

    if (scheme == "nest"):
        nest = True
    else:
        nest = False

    x = np.random.uniform(low=-1.0, high=1.0, size=1_000_000)
    y = np.random.uniform(low=-1.0, high=1.0, size=1_000_000)
    z = np.random.uniform(low=-1.0, high=1.0, size=1_000_000)

    pix_hpgeom = hpgeom.vector_to_pixel(nside, x, y, z, nest=nest)
    pix_healpy = hp.vec2pix(nside, x, y, z, nest=nest)

    np.testing.assert_array_equal(pix_hpgeom, pix_healpy)


def test_vector_to_pixel_single():
    """Test vector_to_pixel, single pixel."""
    x = 0.5
    y = 0.5
    z = 0.5

    pix_hpgeom = hpgeom.vector_to_pixel(1024, x, y, z)

    assert isinstance(pix_hpgeom, np.int64)



import numpy as np
import pytest

try:
    import healpy as hp
    has_healpy = True
except ImportError:
    has_healpy = False

import hpgeom


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10, 2**15, 2**20, 2**25, 2**29])
@pytest.mark.parametrize("scheme", ["nest", "ring"])
def test_pixel_to_vector(nside, scheme):
    """Test pixel_to_vector."""
    np.random.seed(12345)

    if (scheme == "nest"):
        nest = True
    else:
        nest = False

    pix = np.random.randint(low=0, high=hpgeom.nside_to_npixel(nside) - 1, size=1_000_000, dtype=np.int64)

    x_hpgeom, y_hpgeom, z_hpgeom = hpgeom.pixel_to_vector(nside, pix, nest=nest)
    x_healpy, y_healpy, z_healpy = hp.pix2vec(nside, pix, nest=nest)

    np.testing.assert_array_equal(x_hpgeom, x_healpy)
    np.testing.assert_array_equal(y_hpgeom, y_healpy)
    np.testing.assert_array_equal(z_hpgeom, z_healpy)


@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10, 2**15, 2**20, 2**25, 2**29])
def test_pixel_to_vector_bad_pix(nside):
    """Test pixel_to_angle errors when given bad pixel"""

    with pytest.raises(ValueError, match=r"Pixel value .* out of range"):
        hpgeom.pixel_to_vector(nside, -1)

    with pytest.raises(ValueError, match=r"Pixel value .* out of range"):
        hpgeom.pixel_to_vector(nside, 12*nside*nside)


def test_pixel_to_vector_bad_nside():
    """Test pixel_to_angle errors when given a bad nside."""
    np.random.seed(12345)

    pix = np.random.randint(low=0, high=12*2048*2048-1, size=100, dtype=np.int64)

    with pytest.raises(ValueError, match=r"nside .* must be positive"):
        hpgeom.pixel_to_vector(-10, pix, nest=False)

    with pytest.raises(ValueError, match=r"nside .* must be power of 2"):
        hpgeom.pixel_to_vector(2040, pix, nest=True)

    with pytest.raises(ValueError, match=r"nside .* must not be greater than"):
        hpgeom.pixel_to_vector(2**30, pix, nest=True)



import numpy as np
import pytest
import warnings

import hpgeom

from utils import match_arrays


@pytest.mark.parametrize("nside_radius", [(2**5, 2.0),
                                          (2**10, 1.0),
                                          (2**20, 0.01)])
@pytest.mark.parametrize("lon", [0.0, 90.0, 180.0, 270.0])
@pytest.mark.parametrize("lat", [-45.0, 0.0, 45.0, 90.0, -90.0])
def test_query_ellipse_circle_nest(nside_radius, lon, lat):
    """Test query_ellipse with circles, nest ordering."""
    nside = nside_radius[0]
    radius = nside_radius[1]

    # First, non-inclusive
    pixels_ellipse = hpgeom.query_ellipse(nside, lon, lat, radius, radius, 0.0)
    pixels_circle = hpgeom.query_circle(nside, lon, lat, radius)

    np.testing.assert_array_equal(pixels_ellipse, pixels_circle)

    # Second, inclusive.
    pixels_ellipse = hpgeom.query_ellipse(nside, lon, lat, radius, radius, 0.0, inclusive=True)
    pixels_circle = hpgeom.query_circle(nside, lon, lat, radius, inclusive=True)

    np.testing.assert_array_equal(pixels_ellipse, pixels_circle)


@pytest.mark.parametrize("nside_radius", [(2**5, 2.0),
                                          (2**10, 1.0),
                                          (2**20, 0.01)])
@pytest.mark.parametrize("lon", [0.0, 90.0, 180.0, 270.0])
@pytest.mark.parametrize("lat", [-45.0, 0.0, 45.0, 90.0, -90.0])
def test_query_ellipse_circle_ring(nside_radius, lon, lat):
    """Test query_ellipse with circles, ring ordering."""
    nside = nside_radius[0]
    radius = nside_radius[1]

    # First, non-inclusive
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        pixels_ellipse = hpgeom.query_ellipse(nside, lon, lat, radius, radius, 0.0, nest=False)

    pixels_circle = hpgeom.query_circle(nside, lon, lat, radius, nest=False)

    np.testing.assert_array_equal(pixels_ellipse, pixels_circle)

    # Second, inclusive.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        pixels_ellipse = hpgeom.query_ellipse(
            nside,
            lon,
            lat,
            radius,
            radius,
            0.0,
            inclusive=True,
            nest=False
        )

    pixels_circle = hpgeom.query_circle(nside, lon, lat, radius, inclusive=True, nest=False)

    np.testing.assert_array_equal(pixels_ellipse, pixels_circle)


def _pos_in_ellipse(lon, lat, lon_0, lat_0, major, minor, alpha_deg):
    """Get which positions are in an ellipse.

    Following
    https://math.stackexchange.com/questions/3747965/points-within-an-ellipse-on-the-globe

    The sign of alpha has been reversed from the equations posted there so that it is
    defined as the angle East (clockwise) of North.

    Parameters
    ----------
    lon, lat : `np.ndarray`
    lon_0, lat_0 : `float`
    major, minor : `float`
    alpha_deg : `float`

    Returns
    -------
    indices : `np.ndarray`
    """
    theta, phi = hpgeom.lonlat_to_thetaphi(lon, lat)
    theta_0, phi_0 = hpgeom.lonlat_to_thetaphi(lon_0, lat_0)
    vec = hpgeom.angle_to_vector(lon, lat)

    a = np.deg2rad(major)
    b = np.deg2rad(minor)

    gamma = np.sqrt(a**2. - b**2.)
    alpha = np.deg2rad(alpha_deg)

    F1_x = (np.cos(alpha)*np.sin(gamma)*np.cos(phi_0)*np.cos(theta_0)
            + np.sin(alpha)*np.sin(gamma)*np.sin(phi_0)
            + np.cos(gamma)*np.cos(phi_0)*np.sin(theta_0))
    F2_x = (-np.cos(alpha)*np.sin(gamma)*np.cos(phi_0)*np.cos(theta_0)
            - np.sin(alpha)*np.sin(gamma)*np.sin(phi_0)
            + np.cos(gamma)*np.cos(phi_0)*np.sin(theta_0))
    F1_y = (np.cos(alpha)*np.sin(gamma)*np.sin(phi_0)*np.cos(theta_0)
            - np.sin(alpha)*np.sin(gamma)*np.cos(phi_0)
            + np.cos(gamma)*np.sin(phi_0)*np.sin(theta_0))
    F2_y = (-np.cos(alpha)*np.sin(gamma)*np.sin(phi_0)*np.cos(theta_0)
            + np.sin(alpha)*np.sin(gamma)*np.cos(phi_0)
            + np.cos(gamma)*np.sin(phi_0)*np.sin(theta_0))
    F1_z = np.cos(gamma)*np.cos(theta_0) - np.cos(alpha)*np.sin(gamma)*np.sin(theta_0)
    F2_z = np.cos(gamma)*np.cos(theta_0) + np.cos(alpha)*np.sin(gamma)*np.sin(theta_0)

    cos_d1 = vec[:, 0]*F1_x + vec[:, 1]*F1_y + vec[:, 2]*F1_z
    cos_d2 = vec[:, 0]*F2_x + vec[:, 1]*F2_y + vec[:, 2]*F2_z

    inside = ((np.arccos(cos_d1) + np.arccos(cos_d2)) < 2*a)

    return inside


@pytest.mark.parametrize("nside_major_minor", [(2**5, 2.0, 1.0),
                                               (2**10, 1.0, 0.7),
                                               (2**20, 0.01, 0.009)])
@pytest.mark.parametrize("alpha", [0.0, 45.0, 90.0, 135.0, 180.0, 225.0, 270.0, 315.0])
@pytest.mark.parametrize("lon", [0.0, 90.0, 180.0, 270.0])
@pytest.mark.parametrize("lat", [-45.0, 0.0, 45.0, 90.0, -90.0])
def test_query_ellipse_nest_exclusive(nside_major_minor, alpha, lon, lat):
    """Test query_ellipse with ellipses, nest ordering."""
    nside = nside_major_minor[0]
    major = nside_major_minor[1]
    minor = nside_major_minor[2]

    # First, non-inclusive
    pixels_ellipse = hpgeom.query_ellipse(nside, lon, lat, major, minor, alpha)
    pixels_circle = hpgeom.query_circle(nside, lon, lat, major*1.05)
    lon_pix, lat_pix = hpgeom.pixel_to_angle(nside, pixels_circle)
    cut = _pos_in_ellipse(lon_pix, lat_pix, lon, lat, major, minor, alpha)
    pixels_circle_ellipse = pixels_circle[cut]

    np.testing.assert_array_equal(pixels_ellipse, pixels_circle_ellipse)


# The low-res nside=32 tests inclusive are ... inconclusive ... so we skip them.
@pytest.mark.parametrize("nside_major_minor", [(2**10, 1.0, 0.7),
                                               (2**20, 0.01, 0.009)])
@pytest.mark.parametrize("alpha", [0.0, 45.0, 135.0, 225.0, 315.0])
@pytest.mark.parametrize("lon", [0.0, 90.0, 180.0, 270.0])
@pytest.mark.parametrize("lat", [-45.0, 0.0, 45.0, 90.0, -90.0])
def test_query_ellipse_nest_inclusive(nside_major_minor, alpha, lon, lat):
    """Test query_ellipse with ellipses, nest ordering."""
    nside = nside_major_minor[0]
    major = nside_major_minor[1]
    minor = nside_major_minor[2]

    pixels = hpgeom.query_ellipse(nside, lon, lat, major, minor, alpha)
    pixels_ellipse = hpgeom.query_ellipse(nside, lon, lat, major, minor, alpha, inclusive=True)

    # Ensure all the inner pixels are in the inclusive pixels
    sub1, sub2 = match_arrays(pixels_ellipse, pixels)
    assert sub2.size == pixels.size

    # Look at the boundaries of the pixels, check if any are included.
    pixels_circle = hpgeom.query_circle(nside, lon, lat, major*1.1)
    boundaries_lon, boundaries_lat = hpgeom.boundaries(nside, pixels_circle, step=4)
    cut = _pos_in_ellipse(boundaries_lon.ravel(), boundaries_lat.ravel(), lon, lat, major, minor, alpha)
    test = cut.reshape(boundaries_lon.shape).sum(axis=1)
    pixels_circle_ellipse = pixels_circle[test > 0]

    # Ensure all these pixels are in the inclusive list
    sub1, sub2 = match_arrays(pixels_circle_ellipse, pixels_ellipse)
    assert sub1.size == pixels_circle_ellipse.size


def test_query_ellipse_radians():
    """Test query_ellipse, use lonlat and radians."""
    nside = 1024
    major = 2.0
    minor = 1.0
    alpha = 55.0
    lon = 0.0
    lat = 0.0

    pixels_deg = hpgeom.query_ellipse(nside, lon, lat, major, minor, alpha)
    pixels_rad = hpgeom.query_ellipse(
        nside,
        np.deg2rad(lon),
        np.deg2rad(lat),
        np.deg2rad(major),
        np.deg2rad(minor),
        np.deg2rad(alpha),
        degrees=False
    )

    np.testing.assert_array_equal(pixels_rad, pixels_deg)


def test_query_ellipse_thetaphi():
    """Test query_ellipse, use theta_phi."""
    nside = 1024
    major = 2.0
    minor = 1.0
    alpha = -55.0
    lon = 45.0
    lat = 45.0

    pixels_deg = hpgeom.query_ellipse(nside, lon, lat, major, minor, alpha)
    theta, phi = hpgeom.lonlat_to_thetaphi(lon, lat)
    pixels_thetaphi = hpgeom.query_ellipse(
        nside,
        theta,
        phi,
        np.deg2rad(major),
        np.deg2rad(minor),
        np.deg2rad(alpha),
        lonlat=False
    )

    np.testing.assert_array_equal(pixels_thetaphi, pixels_deg)


@pytest.mark.parametrize("fact", [1, 2, 4, 8])
def test_query_ellipse_fact(fact):
    """Test query_ellipse, use other fact values."""
    nside = 1024
    lon = 90.0
    lat = 20.0
    major = 2.0
    minor = 1.0
    alpha = 75.0

    pixels_ellipse = hpgeom.query_ellipse(nside, lon, lat, major, minor, alpha, inclusive=True, fact=fact)

    # Look at the boundaries of the pixels, check if any are included.
    pixels_circle = hpgeom.query_circle(nside, lon, lat, major*1.1)
    boundaries_lon, boundaries_lat = hpgeom.boundaries(nside, pixels_circle, step=4)
    cut = _pos_in_ellipse(boundaries_lon.ravel(), boundaries_lat.ravel(), lon, lat, major, minor, alpha)
    test = cut.reshape(boundaries_lon.shape).sum(axis=1)
    pixels_circle_ellipse = pixels_circle[test > 0]

    # We don't get 100% overlap, unfortunately, so we have to look above a threshold
    sub1 = np.searchsorted(pixels_ellipse, pixels_circle_ellipse)
    bad, = (sub1 == pixels_ellipse.size).nonzero()
    sub1[bad] = pixels_ellipse.size - 1
    sub2, = (pixels_ellipse[sub1] == pixels_circle_ellipse).nonzero()
    sub1 = sub1[sub2]

    assert sub1.size >= int(0.9*pixels_ellipse.size)
    assert sub1.size >= int(0.9*pixels_circle_ellipse.size)


def test_query_ellipse_badinputs():
    """Test query ellipse with bad inputs."""
    with pytest.raises(ValueError, match=r"lat .* out of range"):
        # Latitude out of range
        hpgeom.query_ellipse(2048, 0.0, 100.0, 1.0, 0.5, 0.0)

    with pytest.raises(ValueError, match=r"lat .* out of range"):
        # Latitude out of range
        hpgeom.query_ellipse(2048, 0.0, -100.0, 1.0, 0.5, 0.0)

    with pytest.raises(ValueError, match=r"colatitude \(theta\) .* out of range"):
        # theta out of range
        hpgeom.query_ellipse(2048, -0.1, 0.0, 0.1, 0.05, 0.0, lonlat=False)

    with pytest.raises(ValueError, match=r"colatitude \(theta\) .* out of range"):
        # theta out of range
        hpgeom.query_ellipse(2048, np.pi + 0.1, 0.0, 0.1, 0.05, 0.0, lonlat=False)

    with pytest.raises(ValueError, match=r"longitude \(phi\) .* out of range"):
        # phi out of range
        hpgeom.query_ellipse(2048, 0.0, -0.1, 0.1, 0.05, 0.0, lonlat=False)

    with pytest.raises(ValueError, match=r"longitude \(phi\) .* out of range"):
        # phi out of range
        hpgeom.query_ellipse(2048, 0.0, 2*np.pi + 0.1, 0.1, 0.05, 0.0, lonlat=False)

    with pytest.raises(ValueError, match=r"Semi-major axis must be positive"):
        # Semi-major out of range
        hpgeom.query_ellipse(2048, 0.0, 0.0, 0.0, 0.1, 0.0)

    with pytest.raises(ValueError, match=r"Semi-minor axis must be positive"):
        # Semi-major out of range
        hpgeom.query_ellipse(2048, 0.0, 0.0, 0.2, 0.0, 0.0)

    with pytest.raises(ValueError, match=r"Semi-major axis must be >= semi-minor axis"):
        # Semi-major out of range
        hpgeom.query_ellipse(2048, 0.0, 0.0, 0.2, 0.3, 0.0)

    with pytest.raises(ValueError, match=r"Inclusive factor .* must be positive"):
        # Illegal fact (must be positive)
        hpgeom.query_ellipse(2048, 0.0, 0.0, 1.0, 0.5, 0.0, inclusive=True, fact=0)

    with pytest.raises(ValueError, match=r"Inclusive factor \* nside must be \<\="):
        # Illegal fact (too large)
        hpgeom.query_ellipse(2**28, 0.0, 0.0, 0.0001, 0.00005, 0.0, inclusive=True, fact=4)

    with pytest.raises(ValueError, match=r"Inclusive factor .* must be power of 2 for nest"):
        # Illegal fact (must be power of 2 for nest)
        hpgeom.query_ellipse(2048, 0.0, 0.0, 1.0, 0.5, 0.0, inclusive=True, fact=3)

    # Different platforms have different strings here, but they all say ``integer``.
    with pytest.raises(TypeError, match=r"integer"):
        # Illegal fact (must be integer)
        hpgeom.query_ellipse(2048, 0.0, 0.0, 1.0, 0.5, 0.0, inclusive=True, nest=False, fact=3.5)

    # Check resource warning
    with pytest.warns(ResourceWarning, match=r"natively supports nest ordering"):
        hpgeom.query_ellipse(1024, 0.0, 1.0, 0.5, 0.5, 0.0, nest=False)



import numpy as np
import pytest

try:
    import healpy as hp
    has_healpy = True
except ImportError:
    has_healpy = False

import hpgeom


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10, 2**15, 2**20, 2**25, 2**29, 2**10 - 100])
def test_pixel_to_angle_ring(nside):
    """Test pixel_to_angle for ring scheme."""
    np.random.seed(12345)

    pix = np.random.randint(low=0, high=12*nside*nside-1, size=1_000_000, dtype=np.int64)

    lon_hpgeom, lat_hpgeom = hpgeom.pixel_to_angle(nside, pix, nest=False, lonlat=True, degrees=True)
    lon_healpy, lat_healpy = hp.pix2ang(nside, pix, nest=False, lonlat=True)

    np.testing.assert_array_almost_equal(lon_hpgeom, lon_healpy)
    np.testing.assert_array_almost_equal(lat_hpgeom, lat_healpy)

    lon_hpgeom, lat_hpgeom = hpgeom.pixel_to_angle(
        nside,
        pix,
        nest=False,
        lonlat=True,
        degrees=False
    )

    np.testing.assert_array_almost_equal(np.rad2deg(lon_hpgeom), lon_healpy)
    np.testing.assert_array_almost_equal(np.rad2deg(lat_hpgeom), lat_healpy)

    theta_hpgeom, phi_hpgeom = hpgeom.pixel_to_angle(nside, pix, nest=False, lonlat=False)
    theta_healpy, phi_healpy = hp.pix2ang(nside, pix, nest=False, lonlat=False)

    np.testing.assert_array_almost_equal(theta_hpgeom, theta_healpy)
    np.testing.assert_array_almost_equal(phi_hpgeom, phi_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10, 2**15, 2**20, 2**25, 2**29])
def test_pixel_to_angle_nest(nside):
    """Test pixel_to_angle for nest scheme."""
    np.random.seed(12345)

    pix = np.random.randint(low=0, high=12*nside*nside-1, size=1_000_000, dtype=np.int64)

    lon_hpgeom, lat_hpgeom = hpgeom.pixel_to_angle(nside, pix, nest=True, lonlat=True, degrees=True)
    lon_healpy, lat_healpy = hp.pix2ang(nside, pix, nest=True, lonlat=True)

    np.testing.assert_array_almost_equal(lon_hpgeom, lon_healpy)
    np.testing.assert_array_almost_equal(lat_hpgeom, lat_healpy)

    lon_hpgeom, lat_hpgeom = hpgeom.pixel_to_angle(
        nside,
        pix,
        nest=True,
        lonlat=True,
        degrees=False
    )

    np.testing.assert_array_almost_equal(np.rad2deg(lon_hpgeom), lon_healpy)
    np.testing.assert_array_almost_equal(np.rad2deg(lat_hpgeom), lat_healpy)

    theta_hpgeom, phi_hpgeom = hpgeom.pixel_to_angle(nside, pix, nest=True, lonlat=False)
    theta_healpy, phi_healpy = hp.pix2ang(nside, pix, nest=True, lonlat=False)

    np.testing.assert_array_almost_equal(theta_hpgeom, theta_healpy)
    np.testing.assert_array_almost_equal(phi_hpgeom, phi_healpy)


@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10, 2**15, 2**20, 2**25, 2**29])
def test_pixel_to_angle_scalar(nside):
    """Test pixel_to_angle for scalars."""
    np.random.seed(12345)

    pix = np.random.randint(low=0, high=12*nside*nside-1, size=100, dtype=np.int64)

    lon_arr, lat_arr = hpgeom.pixel_to_angle(nside, pix, nest=True, lonlat=True, degrees=True)
    lon_scalar1, lat_scalar1 = hpgeom.pixel_to_angle(nside, pix[0], nest=True, lonlat=True, degrees=True)

    assert lon_scalar1 == lon_arr[0]
    assert lat_scalar1 == lat_arr[0]

    lon_scalar2, lat_scalar2 = hpgeom.pixel_to_angle(
        nside,
        int(pix[0]),
        nest=True,
        lonlat=True,
        degrees=True
    )

    assert lon_scalar2 == lon_arr[0]
    assert lat_scalar2 == lat_arr[0]


@pytest.mark.parametrize("nside", [2**0, 2**5, 2**10, 2**15, 2**20, 2**25, 2**29])
def test_pixel_to_angle_bad_pix(nside):
    """Test pixel_to_angle errors when given bad pixel"""

    with pytest.raises(ValueError, match=r"Pixel value .* out of range"):
        hpgeom.pixel_to_angle(nside, -1)

    with pytest.raises(ValueError, match=r"Pixel value .* out of range"):
        hpgeom.pixel_to_angle(nside, 12*nside*nside)


def test_pixel_to_angle_bad_nside():
    """Test pixel_to_angle errors when given a bad nside."""
    np.random.seed(12345)

    pix = np.random.randint(low=0, high=12*2048*2048-1, size=100, dtype=np.int64)

    with pytest.raises(ValueError, match=r"nside .* must be positive"):
        hpgeom.pixel_to_angle(-10, pix, nest=False)

    with pytest.raises(ValueError, match=r"nside .* must be positive"):
        hpgeom.pixel_to_angle(-10, pix, nest=True)

    with pytest.raises(ValueError, match=r"nside .* must be power of 2"):
        hpgeom.pixel_to_angle(2040, pix, nest=True)

    with pytest.raises(ValueError, match=r"nside .* must not be greater"):
        hpgeom.pixel_to_angle(2**30, pix, nest=True)



from setuptools import setup, Extension
import numpy


ext = Extension(
    "hpgeom._hpgeom",
    [
        "hpgeom/hpgeom_stack.c",
        "hpgeom/hpgeom_utils.c",
        "hpgeom/healpix_geom.c",
        "hpgeom/hpgeom.c",
    ],
)

setup(
    ext_modules=[ext],
    include_dirs=numpy.get_include(),
)



import warnings

from ._hpgeom import (
    angle_to_pixel,
    pixel_to_angle,
    nest_to_ring,
    ring_to_nest,
    neighbors,
    vector_to_pixel,
    pixel_to_vector,
    max_pixel_radius,
)
from .hpgeom import (
    query_circle_vec,
    query_polygon_vec,
    nside_to_npixel,
    npixel_to_nside,
    nside_to_pixel_area,
    nside_to_resolution,
    nside_to_order,
    order_to_nside,
    angle_to_vector,
    vector_to_angle,
    UNSEEN,
)

from .hpgeom import boundaries as hpgeom_boundaries

__all__ = [
    'ang2pix',
    'pix2ang',
    'query_disc',
    'query_polygon',
    'ring2nest',
    'nest2ring',
    'nside2npix',
    'npix2nside',
    'nside2pixarea',
    'nside2resol',
    'nside2order',
    'order2nside',
    'ang2vec',
    'vec2ang',
    'pix2vec',
    'vec2pix',
    'boundaries',
    'get_all_neighbours',
    'max_pixrad',
    'UNSEEN',
]


def ang2pix(nside, theta, phi, nest=False, lonlat=False):
    """Convert angles to pixels, with healpy interface.

    Parameters
    ----------
    nside : `int`
        The HEALPix nside parameter.
    theta : `np.ndarray` (N,) or scalar
        Angular coordinate theta (radians) or longitude (degrees).
    phi : `np.ndarray` (N,) or scalar
        Angular coordinate phi (radians) or latitude (degrees).
    nest : `bool`, optional
        If True, NEST pixel ordering scheme, otherwise RING.
    lonlat : `bool`, optional
        If True, input angles are lon/lat instead of theta/phi.

    Returns
    -------
    pix : `np.ndarray` (N,) or int
        HEALPix pixel numbers.
    """
    return angle_to_pixel(nside, theta, phi, nest=nest, lonlat=lonlat)


def pix2ang(nside, pix, nest=False, lonlat=False):
    """Convert pixels to angles, with healpy interface.

    Parameters
    ----------
    nside : `int`
        The HEALPix nside parameter.
    pix : `np.ndarrat` (N,) or scalar
        Pixel numbers.
    nest : `bool`, optional
        If True, NEST pixel ordering scheme, otherwise RING.
    lonlat : `bool`, optional
        If True, input angles are lon/lat instead of theta/phi.

    Returns
    -------
    theta : `np.ndarray` (N,) or scalar
        Angular coordinate theta (radians) or longitude (degrees).
    phi : `np.ndarray` (N,) or scalar
        Angular coordinate phi (radians) or latitude (degrees).
    """
    return pixel_to_angle(nside, pix, nest=nest, lonlat=lonlat)


def query_disc(nside, vec, radius, inclusive=False, fact=4, nest=False, buff=None):
    """Returns pixels whose centers lie within the disk defined by
    *vec* and *radius* (in radians) (if *inclusive* is False), or which
    overlap with this disk (if *inclusive* is True).

    Parameters
    ----------
    nside : `int`
        The HEALPix nside parameter.
    vec : float, sequence of 3 elements
        The coordinates of unit vector defining the disk center.
    radius : `float`
        The radius (in radians) of the disk
    inclusive : `bool`, optional
        If False, return the exact set of pixels whose pixel centers lie
        within the disk; if True, return all pixels that overlap with the disk,
        and maybe a few more.
    fact : `int`, optional
        Only used when inclusive=True. The overlapping test will be done at
        the resolution fact*nside. For NESTED ordering, fact must be a power of 2, less than 2**30,
        else it can be any positive integer.
    nest: `bool`, optional
        If True, assume NESTED pixel ordering, otherwise, RING pixel ordering
    buff: `int` array, optional
        If provided, this numpy array is used to contain the return values and must be
        at least long enough to do so.  Note that this is provided for compatibility
        only and is not optimized.

    Returns
    -------
    pixels : `np.ndarray` (N,)
        Array of pixels (`np.int64`) which cover the disk.
    """
    pixels = query_circle_vec(nside, vec, radius, inclusive=inclusive, fact=fact, nest=nest)

    if buff is not None:
        warnings.warn("In hpgeom, setting buff is less performant than simply returning the pixels.")
        buff[0: len(pixels)] = pixels

    return pixels


def query_polygon(nside, vertices, inclusive=False, fact=4, nest=False, buff=None):
    """Returns pixels whose centers lie within the polygon defined by
    *vertices* (if *inclusive* is False), or which overlap with the polygon
    (if *inclusive* is True).

    Parameters
    ----------
    nside : `int`
        The HEALPix nside parameter.
    vertices : `np.ndarray` (N, 3)
        Vertex array containing the vertices of the polygon.
    inclusive : `bool`, optional
        If False, return the exact set of pixels whose pixel centers lie
        within the polygon; if True, return all pixels that overlap with the polygon,
        and maybe a few more.
    fact : `int`, optional
        Only used when inclusive=True. The overlapping test will be done at
        the resolution fact*nside. For NESTED ordering, fact must be a power of 2, less than 2**30,
        else it can be any positive integer.
    nest: `bool`, optional
        If True, assume NESTED pixel ordering, otherwise, RING pixel ordering
    buff: `int` array, optional
        If provided, this numpy array is used to contain the return values and must be
        at least long enough to do so.  Note that this is provided for compatibility
        only and is not optimized.

    Returns
    -------
    pixels : `np.ndarray` (N,)
        Array of pixels (`np.int64`) which cover the disk.
    """
    pixels = query_polygon_vec(nside, vertices, inclusive=inclusive, fact=fact, nest=nest)

    if buff is not None:
        warnings.warn("In hpgeom, setting buff is less performant than simply returning the pixels.")
        buff[0: len(pixels)] = pixels

    return pixels


def nest2ring(nside, pix):
    """Convert pixel number from nest to ring ordering.

    Parameters
    ----------
    nside : `int`, scalar
        The healpix nside parameter.  Must be power of 2.
    pix : `int` or `np.ndarray` (N,)
        The pixel numbers in nest scheme.

    Returns
    -------
    pix : `int` or `np.ndarray` (N,)
        The pixel numbers in ring scheme.
    """
    return nest_to_ring(nside, pix)


def ring2nest(nside, pix):
    """Convert pixel number from ring to nest ordering.

    Parameters
    ----------
    nside : `int`, scalar
        The healpix nside parameter.  Must be power of 2.
    pix : `int` or `np.ndarray` (N,)
        The pixel numbers in ring scheme.

    Returns
    -------
    pix : `int` or `np.ndarray` (N,)
        The pixel numbers in nest scheme.
    """
    return ring_to_nest(nside, pix)


def nside2npix(nside):
    """Return the number of pixels given an nside.

    Parameters
    ----------
    nside : `int` or `np.ndarray` (N,)
        HEALPix nside

    Returns
    -------
    npixel : `int` or `np.ndarray` (N,)
        Number of pixels associated with that nside.
    """
    return nside_to_npixel(nside)


def npix2nside(nside):
    """Return the nside given a number of pixels.

    Parameters
    ----------
    npixel : `int` or `np.ndarray` (N,)
        Number of pixels.

    Returns
    -------
    nside : `int` or `np.ndarray` (N,)
        HEALPix nside associated with that number of pixels.
    """
    return npixel_to_nside(nside)


def nside2pixarea(nside, degrees=False):
    """Return the pixel area given an nside in square degrees or square radians.

    Parameters
    ----------
    nside : `int`
        HEALPix nside parameter.
    degrees : `bool`, optional
        Return area in square degrees?

    Returns
    -------
    pixel_area : `float`
        Pixel area in square degrees or square radians.
    """
    return nside_to_pixel_area(nside, degrees=degrees)


def nside2resol(nside, arcmin=False):
    """Return the approximate resolution (pixel size in radians or arcminutes),
    given an nside.

    Resolution is just the square root of the pixel area, which is an approximation
    given the varying pixel shapes.

    Parameters
    ----------
    nside : `int`
        HEALPix nside parameter.
    arcmin : `bool`, optional
       If True, return resolution in arcminutes, otherwise radians.

    Returns
    -------
    resolution : `float`
        Approximate pixel size in specified units.
    """
    if arcmin:
        units = 'arcminutes'
    else:
        units = 'radians'
    return nside_to_resolution(nside, units=units)


def nside2order(nside):
    """Return the resolution order for a given nside.

    Parameters
    ----------
    nside : `int`
        HEALPix nside parameter.  Will raise ValueError if nside is not valid
        (must be a power of 2 and less than 2**30).

    Returns
    -------
    order : `int`
        Order corresponding to given nside, such that nside = 2**order.
    """
    return nside_to_order(nside)


def order2nside(order):
    """Return the nside for a given order.

    Parameters
    ----------
    order : `int`
        Resolution order.  Will raise ValueError if order is not valid
        (must be 0 to 29 inclusive).

    Returns
    -------
    nside : `int`
        HEALPix nside corresponding to given order, such that nside = 2**order.
    """
    return order_to_nside(order)


def ang2vec(theta, phi, lonlat=False):
    """Convert angles to cartesian unit vectors.

    Parameters
    ----------
    theta : `float` or `np.ndarray` (N,)
        Co-latitude (theta, radians) or longitude (degrees).
    phi : `float` or `np.ndarray` (N,)
        Longitude (phi, radians) or latitude (degrees).
    lonlat : `bool`, optional
        Use longitude/latitude (degrees) instead of longitude/co-latitude (radians).

    Returns
    -------
    vec : `np.ndarray` (N, 3) or (3,)
        If theta, phi are vectors, returns a 2D array with dimensions (N, 3) with
        one vector per row.  If not vectors, a 1D array with dimensions (3,).
    """
    return angle_to_vector(theta, phi, lonlat=lonlat, degrees=True)


def vec2ang(vectors, lonlat=False):
    """Convert cartesian (x, y, z) vectors to angles.

    Parameters
    ----------
    vec : `np.ndarray` (3,) or (N, 3)
        The vectors to convert to angles.
    lonlat : `bool`, optional
        Use longitude/latitude instead of co-latitude/longitude (radians).
    degrees : `bool`, optional
        If lonlat is True then this sets if the units are degrees or
        radians.

    Returns
    -------
    theta : `float` or `np.ndarray` (N,)
        Longitude or co-latitude theta (radians if lonlat=False, degrees if
        lonlat=True and degrees=True).
    phi : `float` or `np.ndarray` (N,)
        Latitude or longitude phi (radians if lonlat=False, degrees if lonlat=True
        and degrees=True).
    """
    return vector_to_angle(vectors, lonlat=lonlat, degrees=True)


def boundaries(nside, pix, step=1, nest=False):
    """Returns an array containing vectors to the boundary of the
    given pixel.

    The returned array has shape (3, 4*step) (for a single pixel) or
    (N, 3, 4*step) for multiple pixels.  The elements are the x,y,z
    positions on the unit sphere of the pixel boundary. To retrieve
    corners, specify step=1.

    Parameters
    ----------
    nside : `int` or `np.ndarray` (N,)
        HEALPix nside.  Must be power of 2 for nest ordering.
    pix : `int` or `np.ndarray` (N,)
        Pixel number(s).
    step : `int`, optional
        Number of steps for each side of the pixel.
    nest : `bool`, optional
        Use nest ordering scheme?

    Returns
    -------
    boundary : `np.ndarray` (N, 3, 4*step)
        x,y,z for positions on the boundary of the pixel.
    """
    theta, phi = hpgeom_boundaries(nside, pix, step=step, nest=nest)

    if theta.ndim == 1:
        # Single pixel
        return angle_to_vector(theta, phi).transpose()
    else:
        # Multiple pixels
        return angle_to_vector(theta, phi).transpose([1, 2, 0])


def pix2vec(nside, pix, nest=False):
    """Convert pixels to cartesian vectors (x, y, z).

    Parameters
    ----------
    nside : `int` or `np.ndarray` (N,)
        HEALPix nside.  Must be power of 2 for nest ordering.
    pix : `int` or `np.ndarray` (N,)
        Pixel number(s).
    nest : `bool`, optional
        Use nest ordering scheme?

    Returns
    -------
    x : `np.ndarray` (N,)
        x coordinates of vectors.
    y : `np.ndarray` (N,)
        y coordinates of vectors.
    z : `np.ndarray` (N,)
        z coordinates of vectors.
    """
    return pixel_to_vector(nside, pix, nest=nest)


def vec2pix(nside, x, y, z, nest=False):
    """Convert cartesian vectors (x, y, z) to pixels.

    Parameters
    ----------
    nside : `int` or `np.ndarray` (N,)
        HEALPix nside.  Must be power of 2 for nest ordering.
    x : `np.ndarray` (N,)
        x coordinates of vectors.
    y : `np.ndarray` (N,)
        y coordinates of vectors.
    z : `np.ndarray` (N,)
        z coordinates of vectors.
    nest : `bool`, optional
        Use nest ordering scheme?

    Returns
    -------
    pix : `int` or `np.ndarray` (N,)
        Pixel number(s).
    """
    return vector_to_pixel(nside, x, y, z, nest=nest)


def get_all_neighbours(nside, theta, phi=None, nest=False, lonlat=False):
    """Return the 8 nearest neighbor pixels.

    Parameters
    ----------
    nside : `int`
        HEALPix nside. Must be power of 2 for nest ordering.
    theta : `np.ndarray` (N,) or `float` or `int`
        If phi is None, this is interpreted as integer pixel numbers.
        If phi is not None, then this is interpreted as angular coordinate
        theta (if lonlat=False) or longitude (if lonlat=True).
    phi : `np.ndarray` (N,) or `float`, optional
        Angular coordinate phi (if lonlat=False) or latitude (if lonlat=True).
    nest : `bool`, optional
        Use nest ordering scheme?
    lonlat : `bool`, optional
        Use longitude/latitude (degrees) instead of co-latitude/longitude (radians)

    Returns
    -------
    pixels : `np.ndarray` (8,) or (8, N)
        Neighboring pixel numbers of the SW, W, NW, N, NE, E, SE, and S neighbors.
        If a neighbor does not exist (as can be the case for W, N, E, and S), the
        corresponding pixel number will be -1.
    """
    if phi is not None:
        _pix = angle_to_pixel(nside, theta, phi, nest=nest, lonlat=lonlat, degrees=True)
    else:
        _pix = theta

    neigh = neighbors(nside, _pix, nest=nest)

    return neigh.transpose()


def max_pixrad(nside, degrees=False):
    """Compute maximum angular distance between any pixel and its corners.

    Parameters
    ----------
    nside : `int` or `np.ndarray`
        HEALPix nside.
    degrees : `bool`, optional
        If True, returns pixel radius in degrees, otherwise radians.

    Returns
    -------
    radii : `np.ndarray` (N, ) or `float`
        Angular distance(s) in degrees or radians.
    """
    return max_pixel_radius(nside, degrees=degrees)



import numpy as np
import time

try:
    import healpy as hp
    has_healpy = True
except ImportError:
    has_healpy = False

import hpgeom


ntrial = 10
np.random.seed(12345)

with open('benchmark_table.md', 'w') as bt:
    bt.write("|Function|Scheme|Nside|Size|Time (healpy)|Time (hpgeom)|hpgeom/healpy|\n")
    bt.write("|--------|------|-----|----|-------------|-------------|-------------|\n")

    function = 'angle_to_pixel'

    for scheme in ['nest', 'ring']:
        if scheme == 'nest':
            nest = True
        else:
            nest = False

        for size in [10, 10_000, 1_000_000]:
            lon = np.random.uniform(low=0.0, high=360.0, size=size)
            lat = np.random.uniform(low=-90.0, high=90.0, size=size)

            for nside in [128, 4096, 2**17]:
                start_time = time.time()
                for i in range(ntrial):
                    _ = hpgeom.angle_to_pixel(nside, lon, lat, nest=nest, lonlat=True)
                time_hpgeom = (time.time() - start_time)/ntrial

                start_time = time.time()
                for i in range(ntrial):
                    _ = hp.ang2pix(nside, lon, lat, nest=nest, lonlat=True)
                time_healpy = (time.time() - start_time)/ntrial

                time_ratio = time_hpgeom/time_healpy

                bt.write(f'|{function}|{scheme}|{nside}|{size}|{time_healpy}|{time_hpgeom}|{time_ratio}|\n')

    function = 'pixel_to_angle'

    for scheme in ['nest', 'ring']:
        if scheme == 'nest':
            nest = True
        else:
            nest = False

        for size in [10, 10_000, 1_000_000]:
            for nside in [128, 4096, 2**17]:
                pix = np.random.randint(low=0, high=12*nside*nside-1, size=size)

                start_time = time.time()
                for i in range(ntrial):
                    _ = hpgeom.pixel_to_angle(nside, pix, nest=nest, lonlat=True)
                time_hpgeom = (time.time() - start_time)/ntrial

                start_time = time.time()
                for i in range(ntrial):
                    _ = hp.pix2ang(nside, pix, nest=nest, lonlat=True)
                time_healpy = (time.time() - start_time)/ntrial

                time_ratio = time_hpgeom/time_healpy

                bt.write(f'|{function}|{scheme}|{nside}|{size}|{time_healpy}|{time_hpgeom}|{time_ratio}|\n')

    function = 'nest_to_ring'

    for scheme in ['nest']:
        for size in [10, 10_000, 1_000_000]:
            for nside in [128, 4096, 2**17]:
                pix = np.random.randint(low=0, high=12*nside*nside-1, size=size)

                start_time = time.time()
                for i in range(ntrial):
                    _ = hpgeom.nest_to_ring(nside, pix)
                time_hpgeom = (time.time() - start_time)/ntrial

                start_time = time.time()
                for i in range(ntrial):
                    _ = hp.nest2ring(nside, pix)
                time_healpy = (time.time() - start_time)/ntrial

                time_ratio = time_hpgeom/time_healpy

                bt.write(f'|{function}|{scheme}|{nside}|{size}|{time_healpy}|{time_hpgeom}|{time_ratio}|\n')

    function = 'ring_to_nest'

    for scheme in ['ring']:
        for size in [10, 10_000, 1_000_000]:
            for nside in [128, 4096, 2**17]:
                pix = np.random.randint(low=0, high=12*nside*nside-1, size=size)

                start_time = time.time()
                for i in range(ntrial):
                    _ = hpgeom.ring_to_nest(nside, pix)
                time_hpgeom = (time.time() - start_time)/ntrial

                start_time = time.time()
                for i in range(ntrial):
                    _ = hp.ring2nest(nside, pix)
                time_healpy = (time.time() - start_time)/ntrial

                time_ratio = time_hpgeom/time_healpy

                bt.write(f'|{function}|{scheme}|{nside}|{size}|{time_healpy}|{time_hpgeom}|{time_ratio}|\n')

    bt.write("\n")
    bt.write("|Function|Scheme|Nside|Radius|Time (healpy)|Time (hpgeom)|hpgeom/healpy|\n")
    bt.write("|--------|------|-----|----|-------------|-------------|-------------|\n")

    function = 'query_circle'

    for scheme in ['nest', 'ring', 'ring2nest']:
        if scheme == 'nest':
            nest = True
        else:
            nest = False

        for radius in np.deg2rad([0.1, 0.5, 1.0]):
            rad_deg = np.rad2deg(radius)
            for nside in [128, 1024, 4096]:
                start_time = time.time()
                for i in range(ntrial):
                    pixels = hpgeom.query_circle(
                        nside,
                        0.0,
                        0.0,
                        radius,
                        nest=nest,
                        lonlat=True,
                        degrees=False
                    )
                    if scheme == 'ring2nest':
                        _ = hpgeom.ring_to_nest(nside, pixels)
                time_hpgeom = (time.time() - start_time)/ntrial

                vec = hp.ang2vec(0.0, 0.0, lonlat=True)
                start_time = time.time()
                for i in range(ntrial):
                    pixels = hp.query_disc(nside, vec, radius, nest=nest)
                    if scheme == 'ring2nest':
                        _ = hp.ring2nest(nside, pixels)
                time_healpy = (time.time() - start_time)/ntrial

                time_ratio = time_hpgeom/time_healpy

                bt.write(f'|{function}|{scheme}|{nside}|{rad_deg}|'
                         f'{time_healpy}|{time_hpgeom}|{time_ratio}|\n')



import numpy as np
import pytest

try:
    import healpy as hp
    has_healpy = True
except ImportError:
    has_healpy = False

import hpgeom
import hpgeom.healpy_compat as hpc


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_ang2pix():
    """Test hpgeom.healpy_compat.ang2pix."""
    np.random.seed(12345)

    nside = 2048

    lon = np.random.uniform(low=0.0, high=360.0, size=1_000_000)
    lat = np.random.uniform(low=-90.0, high=90.0, size=1_000_000)
    theta, phi = hpgeom.lonlat_to_thetaphi(lon, lat)

    pix_hpcompat = hpc.ang2pix(nside, theta, phi)
    pix_healpy = hp.ang2pix(nside, theta, phi)
    np.testing.assert_array_equal(pix_hpcompat, pix_healpy)

    pix_hpcompat = hpc.ang2pix(nside, theta, phi, nest=True)
    pix_healpy = hp.ang2pix(nside, theta, phi, nest=True)
    np.testing.assert_array_equal(pix_hpcompat, pix_healpy)

    pix_hpcompat = hpc.ang2pix(nside, lon, lat, lonlat=True)
    pix_healpy = hpc.ang2pix(nside, lon, lat, lonlat=True)
    np.testing.assert_array_equal(pix_hpcompat, pix_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_pix2ang():
    """Test hpgeom.healpy_compat.pix2ang."""
    np.random.seed(12345)

    nside = 2048

    pix = np.random.randint(low=0, high=12*nside*nside-1, size=1_000_000, dtype=np.int64)

    theta_hpcompat, phi_hpcompat = hpc.pix2ang(nside, pix)
    theta_healpy, phi_healpy = hp.pix2ang(nside, pix)
    np.testing.assert_array_almost_equal(theta_hpcompat, theta_healpy)
    np.testing.assert_array_almost_equal(phi_hpcompat, phi_healpy)

    theta_hpcompat, phi_hpcompat = hpc.pix2ang(nside, pix, nest=True)
    theta_healpy, phi_healpy = hp.pix2ang(nside, pix, nest=True)
    np.testing.assert_array_almost_equal(theta_hpcompat, theta_healpy)
    np.testing.assert_array_almost_equal(phi_hpcompat, phi_healpy)

    lon_hpcompat, lat_hpcompat = hpc.pix2ang(nside, pix, lonlat=True)
    lon_healpy, lat_healpy = hp.pix2ang(nside, pix, lonlat=True)
    np.testing.assert_array_almost_equal(lon_hpcompat, lon_healpy)
    np.testing.assert_array_almost_equal(lat_hpcompat, lat_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_query_disc():
    """Test hpgeom.healpy_compat.query_disc."""
    np.random.seed(12345)

    nside = 2048
    lon = 10.0
    lat = 20.0
    radius = 0.5

    theta, phi = hpgeom.lonlat_to_thetaphi(lon, lat)
    sintheta = np.sin(theta)
    vec = [sintheta*np.cos(phi), sintheta*np.sin(phi), np.cos(theta)]

    pix_hpcompat = hpc.query_disc(nside, vec, radius)
    pix_healpy = hp.query_disc(nside, vec, radius)
    np.testing.assert_array_equal(pix_hpcompat, pix_healpy)

    pix_hpcompat = hpc.query_disc(nside, vec, radius, nest=True)
    pix_healpy = hp.query_disc(nside, vec, radius, nest=True)
    np.testing.assert_array_equal(pix_hpcompat, pix_healpy)

    pix_hpcompat = hpc.query_disc(nside, vec, radius, inclusive=True)
    pix_healpy = hp.query_disc(nside, vec, radius, inclusive=True)
    np.testing.assert_array_equal(pix_hpcompat, pix_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_ring2nest():
    """Test hpgeom.healpy_compat.ring2nest."""
    pix_hpcompat = hpc.ring2nest(2048, 1000)
    pix_healpy = hp.ring2nest(2048, 1000)

    assert pix_hpcompat == pix_healpy

    pix_hpcompat = hpc.ring2nest(2048, np.arange(100))
    pix_healpy = hp.ring2nest(2048, np.arange(100))

    np.testing.assert_array_equal(pix_hpcompat, pix_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_nest2ring():
    """Test hpgeom.healpy_compat.nest2ring."""
    pix_hpcompat = hpc.nest2ring(2048, 1000)
    pix_healpy = hp.nest2ring(2048, 1000)

    assert pix_hpcompat == pix_healpy

    pix_hpcompat = hpc.nest2ring(2048, np.arange(100))
    pix_healpy = hp.nest2ring(2048, np.arange(100))

    np.testing.assert_array_equal(pix_hpcompat, pix_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_nside2npix():
    """Test hpgeom.healpy_compat.nside2npix."""
    npix_hpcompat = hpc.nside2npix(2048)
    npix_healpy = hp.nside2npix(2048)

    assert npix_hpcompat == npix_healpy


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_npix2nside():
    """Test hpgeom.healpy_compat.npix2nside."""
    nside_hpcompat = hpc.nside2npix(12*2048*2048)
    nside_healpy = hp.nside2npix(12*2048*2048)

    assert nside_hpcompat == nside_healpy

    with pytest.raises(ValueError, match=r"Illegal npixel"):
        hpc.npix2nside(100)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_nside2pixarea():
    """Test hpgeom.healpy_compat.nside2pixarea."""
    pixarea_hpcompat = hpc.nside2pixarea(1024)
    pixarea_healpy = hp.nside2pixarea(1024)

    assert pixarea_hpcompat == pixarea_healpy

    pixarea_hpcompat = hpc.nside2pixarea(1024, degrees=True)
    pixarea_healpy = hp.nside2pixarea(1024, degrees=True)

    assert pixarea_hpcompat == pixarea_healpy


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_nside2resol():
    """Test hpgeom.healpy_compat.nside2resol."""
    resol_hpcompat = hpc.nside2resol(1024)
    resol_healpy = hp.nside2resol(1024)

    assert resol_hpcompat == resol_healpy

    resol_hpcompat = hpc.nside2resol(1024, arcmin=True)
    resol_healpy = hp.nside2resol(1024, arcmin=True)

    assert resol_hpcompat == resol_healpy


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_nside2order():
    """Test hpgeom.healpy_compat.nside2order."""
    order_hpcompat = hpc.nside2order(1024)
    order_healpy = hp.nside2order(1024)

    assert (order_hpcompat == order_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_order2nside():
    """Test hpgeom.healpy_compat.order2nside."""
    nside_hpcompat = hpc.order2nside(10)
    nside_healpy = hp.order2nside(10)

    assert (nside_hpcompat == nside_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_ang2vec():
    """Test hpgeom.healpy_compat.ang2vec."""
    np.random.seed(12345)

    lon = np.random.uniform(0, 360.0, 1000)
    lat = np.random.uniform(-90.0, 90.0, 1000)

    vec_hpgeom = hpc.ang2vec(lon, lat, lonlat=True)
    vec_healpy = hp.ang2vec(lon, lat, lonlat=True)

    np.testing.assert_array_almost_equal(vec_hpgeom, vec_healpy)

    theta, phi = hpgeom.lonlat_to_thetaphi(lon, lat)

    vec_hpgeom = hpc.ang2vec(theta, phi)
    vec_healpy = hp.ang2vec(theta, phi)

    np.testing.assert_array_almost_equal(vec_hpgeom, vec_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_vec2ang():
    """Test hpgeom.healpy_compat.vec2ang."""
    np.random.seed(12345)

    vec = np.zeros((1000, 3))
    vec[:, 0] = np.random.uniform(-1, 1, 1000)
    vec[:, 1] = np.random.uniform(-1, 1, 1000)
    vec[:, 2] = np.random.uniform(-1, 1, 1000)

    theta_hpgeom, phi_hpgeom = hpc.vec2ang(vec)
    theta_healpy, phi_healpy = hp.vec2ang(vec)

    np.testing.assert_array_almost_equal(theta_hpgeom, theta_healpy)
    np.testing.assert_array_almost_equal(phi_hpgeom, phi_healpy)

    lon_hpgeom, lat_hpgeom = hpc.vec2ang(vec, lonlat=True)
    lon_healpy, lat_healpy = hp.vec2ang(vec, lonlat=True)

    np.testing.assert_array_almost_equal(lon_hpgeom, lon_healpy)
    np.testing.assert_array_almost_equal(lat_hpgeom, lat_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_vec2pix():
    """Test hpgeom.healpy_compat.vec2pix."""
    np.random.seed(12345)

    nside = 2048

    vec = np.zeros((1000, 3))
    vec[:, 0] = np.random.uniform(-1, 1, 1000)
    vec[:, 1] = np.random.uniform(-1, 1, 1000)
    vec[:, 2] = np.random.uniform(-1, 1, 1000)

    pix_hpgeom = hpc.vec2pix(nside, vec[:, 0], vec[:, 1], vec[:, 2])
    pix_healpy = hp.vec2pix(nside, vec[:, 0], vec[:, 1], vec[:, 2])

    np.testing.assert_array_equal(pix_hpgeom, pix_healpy)

    pix_hpgeom = hpc.vec2pix(nside, vec[:, 0], vec[:, 1], vec[:, 2], nest=True)
    pix_healpy = hp.vec2pix(nside, vec[:, 0], vec[:, 1], vec[:, 2], nest=True)

    np.testing.assert_array_equal(pix_hpgeom, pix_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_pix2vec():
    """Test hpgeom.healpy_compat.pix2vec."""
    np.random.seed(12345)

    nside = 2048

    pix = np.random.randint(low=0, high=12*nside*nside-1, size=1000, dtype=np.int64)

    x_hpgeom, y_hpgeom, z_hpgeom = hpc.pix2vec(nside, pix)
    x_healpy, y_healpy, z_healpy = hp.pix2vec(nside, pix)

    np.testing.assert_array_almost_equal(x_hpgeom, x_healpy)
    np.testing.assert_array_almost_equal(y_hpgeom, y_healpy)
    np.testing.assert_array_almost_equal(z_hpgeom, z_healpy)

    x_hpgeom, y_hpgeom, z_hpgeom = hpc.pix2vec(nside, pix, nest=True)
    x_healpy, y_healpy, z_healpy = hp.pix2vec(nside, pix, nest=True)

    np.testing.assert_array_almost_equal(x_hpgeom, x_healpy)
    np.testing.assert_array_almost_equal(y_hpgeom, y_healpy)
    np.testing.assert_array_almost_equal(z_hpgeom, z_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_boundaries():
    """Test hpgeom.healpy_compat.boundaries."""
    # Test single pixel.

    vec_hpgeom = hpc.boundaries(1024, 1000)
    vec_healpy = hp.boundaries(1024, 1000)

    np.testing.assert_array_almost_equal(vec_hpgeom, vec_healpy)

    vec_hpgeom = hpc.boundaries(1024, 1000, step=4)
    vec_healpy = hp.boundaries(1024, 1000, step=4)

    np.testing.assert_array_almost_equal(vec_hpgeom, vec_healpy)

    vec_hpgeom = hpc.boundaries(1024, 1000, nest=False)
    vec_healpy = hp.boundaries(1024, 1000, nest=False)

    np.testing.assert_array_almost_equal(vec_hpgeom, vec_healpy)

    # Test multiple pixels.
    vec_hpgeom = hpc.boundaries(1024, [1000, 1200])
    vec_healpy = hp.boundaries(1024, [1000, 1200])

    np.testing.assert_array_almost_equal(vec_hpgeom, vec_healpy)

    vec_hpgeom = hpc.boundaries(1024, [1000, 1200], step=4)
    vec_healpy = hp.boundaries(1024, [1000, 1200], step=4)

    np.testing.assert_array_almost_equal(vec_hpgeom, vec_healpy)

    vec_hpgeom = hpc.boundaries(1024, [1000, 1200], nest=False)
    vec_healpy = hp.boundaries(1024, [1000, 1200], nest=False)

    np.testing.assert_array_almost_equal(vec_hpgeom, vec_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_get_all_neighbours():
    """Test hpgeom.healpy_compat.get_all_neighbours."""
    # Test single pixel
    neighbors_hpgeom = hpc.get_all_neighbours(1024, 100)
    neighbors_healpy = hp.get_all_neighbours(1024, 100)

    np.testing.assert_array_equal(neighbors_hpgeom, neighbors_healpy)

    # Test multiple pixels
    neighbors_hpgeom = hpc.get_all_neighbours(1024, [100, 200])
    neighbors_healpy = hp.get_all_neighbours(1024, [100, 200])

    np.testing.assert_array_equal(neighbors_hpgeom, neighbors_healpy)

    # Test single theta/phi
    neighbors_hpgeom = hpc.get_all_neighbours(1024, 0.5, phi=0.5)
    neighbors_healpy = hp.get_all_neighbours(1024, 0.5, phi=0.5)

    np.testing.assert_array_equal(neighbors_hpgeom, neighbors_healpy)

    # Test multiple theta/phi
    neighbors_hpgeom = hpc.get_all_neighbours(1024, [0.5, 0.6], phi=[0.5, 0.6])
    neighbors_healpy = hp.get_all_neighbours(1024, [0.5, 0.6], phi=[0.5, 0.6])

    np.testing.assert_array_equal(neighbors_hpgeom, neighbors_healpy)

    # Test single lon/lat
    neighbors_hpgeom = hpc.get_all_neighbours(1024, 0.5, phi=0.5, lonlat=True)
    neighbors_healpy = hp.get_all_neighbours(1024, 0.5, phi=0.5, lonlat=True)

    np.testing.assert_array_equal(neighbors_hpgeom, neighbors_healpy)

    # Test multiple lon/lat
    neighbors_hpgeom = hpc.get_all_neighbours(1024, [0.5, 0.6], phi=[0.5, 0.6], lonlat=True)
    neighbors_healpy = hp.get_all_neighbours(1024, [0.5, 0.6], phi=[0.5, 0.6], lonlat=True)

    np.testing.assert_array_equal(neighbors_hpgeom, neighbors_healpy)

    # Test multiple lon/lat, nest=True
    neighbors_hpgeom = hpc.get_all_neighbours(1024, [0.5, 0.6], phi=[0.5, 0.6], lonlat=True, nest=True)
    neighbors_healpy = hp.get_all_neighbours(1024, [0.5, 0.6], phi=[0.5, 0.6], lonlat=True, nest=True)

    np.testing.assert_array_equal(neighbors_hpgeom, neighbors_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_query_polygon():
    """Test hpgeom.healpy_compat.query_polygon"""
    nside = 1024
    delta = 1.0
    lon_ref = 180.0
    lat_ref = 0.0
    lon = np.array([lon_ref, lon_ref + delta, lon_ref + delta, lon_ref])
    lat = np.array([lat_ref, lat_ref, lat_ref + delta, lat_ref + delta])
    vec = hpgeom.angle_to_vector(lon, lat)

    pix_hpcompat = hpc.query_polygon(nside, vec)
    pix_healpy = hp.query_polygon(nside, vec)
    np.testing.assert_array_equal(pix_hpcompat, pix_healpy)

    pix_hpcompat = hpc.query_polygon(nside, vec, nest=True)
    pix_healpy = hp.query_polygon(nside, vec, nest=True)
    np.testing.assert_array_equal(pix_hpcompat, pix_healpy)

    pix_hpcompat = hpc.query_polygon(nside, vec, inclusive=True)
    pix_healpy = hp.query_polygon(nside, vec, inclusive=True)
    np.testing.assert_array_equal(pix_hpcompat, pix_healpy)


@pytest.mark.skipif(not has_healpy, reason="Skipping test without healpy")
def test_max_pixrad():
    """Test hpgeom.healpy_compat.max_pixrad"""
    nsides = [1024, 2048]

    radii_hpcompat = hpc.max_pixrad(nsides)
    radii_healpy = hp.max_pixrad(nsides)
    np.testing.assert_array_almost_equal(radii_hpcompat, radii_healpy)

    radii_hpcompat = hpc.max_pixrad(nsides, degrees=True)
    radii_healpy = hp.max_pixrad(nsides, degrees=True)
    np.testing.assert_array_almost_equal(radii_hpcompat, radii_healpy)

    radii_hpcompat = hpc.max_pixrad(nsides[0])
    radii_healpy = hp.max_pixrad(nsides[0])
    np.testing.assert_almost_equal(radii_hpcompat, radii_healpy)
